"""
Functions used for CLI color themes.
"""


import logging
import os

from salt.utils.textformat import TextFormat

log = logging.getLogger(__name__)


def get_color_theme(theme):
    """
    Return the color theme to use
    """
    # Keep the heavy lifting out of the module space
    import salt.utils.data
    import salt.utils.files
    import salt.utils.yaml

    if not os.path.isfile(theme):
        log.warning("The named theme %s if not available", theme)

    try:
        with salt.utils.files.fopen(theme, "rb") as fp_:
            colors = salt.utils.data.decode(salt.utils.yaml.safe_load(fp_))
            ret = {}
            for color in colors:
                ret[color] = "\033[{}m".format(colors[color])
            if not isinstance(colors, dict):
                log.warning("The theme file %s is not a dict", theme)
                return {}
            return ret
    except Exception:  # pylint: disable=broad-except
        log.warning("Failed to read the color theme %s", theme)
        return {}


def get_colors(use=True, theme=None):
    """
    Return the colors as an easy to use dict. Pass `False` to deactivate all
    colors by setting them to empty strings. Pass a string containing only the
    name of a single color to be used in place of all colors. Examples:

    .. code-block:: python

        colors = get_colors()  # enable all colors
        no_colors = get_colors(False)  # disable all colors
        red_colors = get_colors('RED')  # set all colors to red

    """

    colors = {
        "BLACK": TextFormat("black"),
        "DARK_GRAY": TextFormat("bold", "black"),
        "RED": TextFormat("red"),
        "LIGHT_RED": TextFormat("bold", "red"),
        "GREEN": TextFormat("green"),
        "LIGHT_GREEN": TextFormat("bold", "green"),
        "YELLOW": TextFormat("yellow"),
        "LIGHT_YELLOW": TextFormat("bold", "yellow"),
        "BLUE": TextFormat("blue"),
        "LIGHT_BLUE": TextFormat("bold", "blue"),
        "MAGENTA": TextFormat("magenta"),
        "LIGHT_MAGENTA": TextFormat("bold", "magenta"),
        "CYAN": TextFormat("cyan"),
        "LIGHT_CYAN": TextFormat("bold", "cyan"),
        "LIGHT_GRAY": TextFormat("white"),
        "WHITE": TextFormat("bold", "white"),
        "DEFAULT_COLOR": TextFormat("default"),
        "ENDC": TextFormat("reset"),
    }
    if theme:
        colors.update(get_color_theme(theme))

    if not use:
        for color in colors:
            colors[color] = ""
    if isinstance(use, str):
        # Try to set all of the colors to the passed color
        if use in colors:
            for color in colors:
                # except for color reset
                if color == "ENDC":
                    continue
                colors[color] = colors[use]

    return colors



"""
unittests for table outputter
"""


import salt.output.table_out as table_out
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class TableTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.output.table_out
    """

    def setup_loader_modules(self):
        return {table_out: {}}

    # The test data should include unicode chars, and in Python 2 there should
    # be an example both of an encoded str type and an actual unicode type.
    # Since unicode_literals is imported, we will achieve the former using
    # salt.utils.stringutils.to_str and the latter by simply using a string
    # literal.
    data = [
        {
            "Food": salt.utils.stringutils.to_str("яйца, бекон, колбаса и спам"),
            "Price": 5.99,
        },
        {"Food": "спам, спам, спам, яйца и спам", "Price": 3.99},
    ]

    def test_output(self):
        ret = table_out.output(self.data)
        self.assertEqual(
            ret,
            "    -----------------------------------------\n"
            "    |              Food             | Price |\n"
            "    -----------------------------------------\n"
            "    |  яйца, бекон, колбаса и спам  |  5.99 |\n"
            "    -----------------------------------------\n"
            "    | спам, спам, спам, яйца и спам |  3.99 |\n"
            "    -----------------------------------------",
        )



"""
Validate the config system
"""

import pytest
from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
class ConfigTest(ModuleCase):
    """
    Test config routines
    """

    @pytest.mark.slow_test
    def test_valid_file_proto(self):
        """
        test config.valid_file_proto
        """
        self.assertTrue(self.run_function("config.valid_fileproto", ["salt://"]))
        self.assertTrue(self.run_function("config.valid_fileproto", ["file://"]))
        self.assertTrue(self.run_function("config.valid_fileproto", ["http://"]))
        self.assertTrue(self.run_function("config.valid_fileproto", ["https://"]))
        self.assertTrue(self.run_function("config.valid_fileproto", ["ftp://"]))
        self.assertTrue(self.run_function("config.valid_fileproto", ["s3://"]))
        self.assertTrue(self.run_function("config.valid_fileproto", ["swift://"]))
        self.assertFalse(self.run_function("config.valid_fileproto", ["cheese://"]))

    @pytest.mark.slow_test
    def test_backup_mode(self):
        """
        test config.backup_mode
        """
        self.assertEqual(self.run_function("config.backup_mode", ["minion"]), "minion")

    @pytest.mark.slow_test
    def test_manage_mode(self):
        """
        test config.manage_mode
        """
        # This function is generally only used with cross calls, the yaml
        # interpreter is breaking it for remote calls
        # The correct standard is the four digit form.
        self.assertEqual(self.run_function("config.manage_mode", ['"775"']), "0775")
        self.assertEqual(self.run_function("config.manage_mode", ['"1775"']), "1775")
        self.assertEqual(self.run_function("config.manage_mode", ['"0775"']), "0775")
        self.assertEqual(self.run_function("config.manage_mode", ['"01775"']), "1775")
        self.assertEqual(self.run_function("config.manage_mode", ['"0"']), "0000")
        self.assertEqual(self.run_function("config.manage_mode", ["775"]), "0775")
        self.assertEqual(self.run_function("config.manage_mode", ["1775"]), "1775")
        self.assertEqual(self.run_function("config.manage_mode", ["0"]), "0000")

    @pytest.mark.slow_test
    def test_option(self):
        """
        test config.option
        """
        # Minion opt
        self.assertEqual(
            self.run_function("config.option", ["master_port"]),
            self.get_config("minion")["master_port"],
        )
        # pillar conf opt
        self.assertEqual(self.run_function("config.option", ["ext_spam"]), "eggs")

    @pytest.mark.slow_test
    def test_get(self):
        """
        Test option.get
        """
        # Check pillar get
        self.assertEqual(self.run_function("config.get", ["level1:level2"]), "foo")
        # Check master config
        self.assertEqual(
            self.run_function("config.get", ["config_opt:layer2"]), "kosher"
        )
        # Check minion config
        self.assertEqual(self.run_function("config.get", ["config_test:spam"]), "eggs")



"""
Tests for the service state
"""

import re

import pytest
import salt.utils.path
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin

INIT_DELAY = 5


@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
class ServiceTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the service state
    """

    def setUp(self):
        self.service_name = "cron"
        cmd_name = "crontab"
        os_family = self.run_function("grains.get", ["os_family"])
        os_release = self.run_function("grains.get", ["osrelease"])
        is_systemd = self.run_function("grains.get", ["systemd"])
        self.stopped = False
        self.running = True
        if os_family == "RedHat":
            self.service_name = "crond"
        elif os_family == "Arch":
            self.service_name = "sshd"
            cmd_name = "systemctl"
        elif os_family == "MacOS":
            self.service_name = "org.ntp.ntpd"
            if int(os_release.split(".")[1]) >= 13:
                self.service_name = "com.apple.AirPlayXPCHelper"
        elif os_family == "Windows":
            self.service_name = "Spooler"

        self.pre_srv_enabled = (
            True
            if self.service_name in self.run_function("service.get_enabled")
            else False
        )
        self.post_srv_disable = False
        if not self.pre_srv_enabled:
            self.run_function("service.enable", name=self.service_name)
            self.post_srv_disable = True

        if os_family != "Windows" and salt.utils.path.which(cmd_name) is None:
            self.skipTest("{} is not installed".format(cmd_name))

        if is_systemd and self.run_function("service.offline"):
            self.skipTest("systemd is OFFLINE")

    def tearDown(self):
        if self.post_srv_disable:
            self.run_function("service.disable", name=self.service_name)

    def check_service_status(self, exp_return):
        """
        helper method to check status of service
        """
        check_status = self.run_function("service.status", name=self.service_name)

        try:
            if not re.match(exp_return, check_status):
                self.fail("status of service is not returning correctly")
        except TypeError:
            if check_status is not exp_return:
                self.fail("status of service is not returning correctly")

    @pytest.mark.slow_test
    def test_service_running(self):
        """
        test service.running state module
        """
        if self.run_function("service.status", name=self.service_name):
            stop_service = self.run_function("service.stop", name=self.service_name)
            self.assertTrue(stop_service)
        self.check_service_status(self.stopped)

        if salt.utils.platform.is_darwin():
            # make sure the service is enabled on macosx
            enable = self.run_function("service.enable", name=self.service_name)

        start_service = self.run_state("service.running", name=self.service_name)
        self.assertTrue(start_service)
        self.check_service_status(self.running)

    @pytest.mark.slow_test
    def test_service_dead(self):
        """
        test service.dead state module
        """
        start_service = self.run_state("service.running", name=self.service_name)
        self.assertSaltTrueReturn(start_service)
        self.check_service_status(self.running)

        ret = self.run_state("service.dead", name=self.service_name)
        self.assertSaltTrueReturn(ret)
        self.check_service_status(self.stopped)

    @pytest.mark.slow_test
    def test_service_dead_init_delay(self):
        """
        test service.dead state module with init_delay arg
        """
        start_service = self.run_state("service.running", name=self.service_name)
        self.assertSaltTrueReturn(start_service)
        self.check_service_status(self.running)

        ret = self.run_state(
            "service.dead", name=self.service_name, init_delay=INIT_DELAY
        )
        self.assertSaltTrueReturn(ret)
        self.check_service_status(self.stopped)



# Import Salt Module
import salt.modules.nginx as nginx
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import Mock, patch
from tests.support.unit import TestCase

MOCK_STATUS_OUTPUT = """Active connections: 7
server accepts handled requests
 46756 46756 89318
Reading: 0 Writing: 7 Waiting: 0"""


class MockUrllibStatus:
    """Mock of urllib2 call for Nginx status"""

    def read(self):
        return MOCK_STATUS_OUTPUT

    def close(self):
        pass


class NginxTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        patcher = patch("salt.utils.path.which", Mock(return_value="/usr/bin/nginx"))
        patcher.start()
        self.addCleanup(patcher.stop)
        return {}

    def test_nginx_status(self):
        mock = Mock(return_value=MockUrllibStatus())
        with patch("urllib.request.urlopen", mock):
            result = nginx.status()
        mock.assert_called_once_with("http://127.0.0.1/status")
        self.assertEqual(
            result,
            {
                "active connections": 7,
                "accepted": 46756,
                "handled": 46756,
                "requests": 89318,
                "reading": 0,
                "writing": 7,
                "waiting": 0,
            },
        )

    def test_nginx_status_with_arg(self):
        mock = Mock(return_value=MockUrllibStatus())
        other_path = "http://localhost/path"
        with patch("urllib.request.urlopen", mock):
            result = nginx.status(other_path)
        mock.assert_called_once_with(other_path)



"""
:maintainer:    Alberto Planas <aplanas@suse.com>
:platform:      Linux
"""


import salt.modules.freezer as freezer
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class FreezerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.freezer
    """

    def setup_loader_modules(self):
        return {freezer: {"__salt__": {}, "__opts__": {"cachedir": ""}}}

    @patch("os.path.isfile")
    def test_status(self, isfile):
        """
        Test if a frozen state exist.
        """
        isfile.side_effect = (True, True)
        self.assertTrue(freezer.status())

        isfile.side_effect = (True, False)
        self.assertFalse(freezer.status())

    @patch("os.listdir")
    @patch("os.path.isdir")
    def test_list(self, isdir, listdir):
        """
        Test the listing of all frozen states.
        """
        # There is no freezer directory
        isdir.return_value = False
        self.assertEqual(freezer.list_(), [])

        # There is freezer directory, but is empty
        isdir.return_value = True
        listdir.return_value = []
        self.assertEqual(freezer.list_(), [])

        # There is freezer directory with states
        isdir.return_value = True
        listdir.return_value = [
            "freezer-pkgs.yml",
            "freezer-reps.yml",
            "state-pkgs.yml",
            "state-reps.yml",
            "random-file",
        ]
        self.assertEqual(freezer.list_(), ["freezer", "state"])

    @patch("os.makedirs")
    def test_freeze_fails_cache(self, makedirs):
        """
        Test to freeze a current installation
        """
        # Fails when creating the freeze cache directory
        makedirs.side_effect = OSError()
        self.assertRaises(CommandExecutionError, freezer.freeze)

    @patch("salt.modules.freezer.status")
    @patch("os.makedirs")
    def test_freeze_fails_already_frozen(self, makedirs, status):
        """
        Test to freeze a current installation
        """
        # Fails when there is already a frozen state
        status.return_value = True
        self.assertRaises(CommandExecutionError, freezer.freeze)
        makedirs.assert_called_once()

    @patch("salt.utils.json.dump")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    @patch("os.makedirs")
    def test_freeze_success_two_freeze(self, makedirs, status, fopen, dump):
        """
        Test to freeze a current installation
        """
        # Freeze the current new state
        status.return_value = False
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={}),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertTrue(freezer.freeze("one"))
            self.assertTrue(freezer.freeze("two"))

            self.assertEqual(makedirs.call_count, 2)
            self.assertEqual(salt_mock["pkg.list_pkgs"].call_count, 2)
            self.assertEqual(salt_mock["pkg.list_repos"].call_count, 2)
            fopen.assert_called()
            dump.assert_called()

    @patch("salt.utils.json.dump")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    @patch("os.makedirs")
    def test_freeze_success_new_state(self, makedirs, status, fopen, dump):
        """
        Test to freeze a current installation
        """
        # Freeze the current new state
        status.return_value = False
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={}),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertTrue(freezer.freeze())
            makedirs.assert_called_once()
            salt_mock["pkg.list_pkgs"].assert_called_once()
            salt_mock["pkg.list_repos"].assert_called_once()
            fopen.assert_called()
            dump.assert_called()

    @patch("salt.utils.json.dump")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    @patch("os.makedirs")
    def test_freeze_success_force(self, makedirs, status, fopen, dump):
        """
        Test to freeze a current installation
        """
        # Freeze the current old state
        status.return_value = True
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={}),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertTrue(freezer.freeze(force=True))
            makedirs.assert_called_once()
            salt_mock["pkg.list_pkgs"].assert_called_once()
            salt_mock["pkg.list_repos"].assert_called_once()
            fopen.assert_called()
            dump.assert_called()

    @patch("salt.modules.freezer.status")
    def test_restore_fails_missing_state(self, status):
        """
        Test to restore an old state
        """
        # Fails if the state is not found
        status.return_value = False
        self.assertRaises(CommandExecutionError, freezer.restore)

    @patch("salt.utils.json.load")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    def test_restore_add_missing_repo(self, status, fopen, load):
        """
        Test to restore an old state
        """
        # Only a missing repo is installed
        status.return_value = True
        load.side_effect = ({}, {"missing-repo": {}})
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={}),
            "pkg.mod_repo": MagicMock(),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertEqual(
                freezer.restore(),
                {
                    "pkgs": {"add": [], "remove": []},
                    "repos": {"add": ["missing-repo"], "remove": []},
                    "comment": [],
                },
            )
            salt_mock["pkg.list_pkgs"].assert_called()
            salt_mock["pkg.list_repos"].assert_called()
            salt_mock["pkg.mod_repo"].assert_called_once()
            fopen.assert_called()
            load.assert_called()

    @patch("salt.utils.json.load")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    def test_restore_add_missing_package(self, status, fopen, load):
        """
        Test to restore an old state
        """
        # Only a missing package is installed
        status.return_value = True
        load.side_effect = ({"missing-package": {}}, {})
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={}),
            "pkg.install": MagicMock(),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertEqual(
                freezer.restore(),
                {
                    "pkgs": {"add": ["missing-package"], "remove": []},
                    "repos": {"add": [], "remove": []},
                    "comment": [],
                },
            )
            salt_mock["pkg.list_pkgs"].assert_called()
            salt_mock["pkg.list_repos"].assert_called()
            salt_mock["pkg.install"].assert_called_once()
            fopen.assert_called()
            load.assert_called()

    @patch("salt.utils.json.load")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    def test_restore_remove_extra_package(self, status, fopen, load):
        """
        Test to restore an old state
        """
        # Only an extra package is removed
        status.return_value = True
        load.side_effect = ({}, {})
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={"extra-package": {}}),
            "pkg.list_repos": MagicMock(return_value={}),
            "pkg.remove": MagicMock(),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertEqual(
                freezer.restore(),
                {
                    "pkgs": {"add": [], "remove": ["extra-package"]},
                    "repos": {"add": [], "remove": []},
                    "comment": [],
                },
            )
            salt_mock["pkg.list_pkgs"].assert_called()
            salt_mock["pkg.list_repos"].assert_called()
            salt_mock["pkg.remove"].assert_called_once()
            fopen.assert_called()
            load.assert_called()

    @patch("salt.utils.json.load")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    def test_restore_remove_extra_repo(self, status, fopen, load):
        """
        Test to restore an old state
        """
        # Only an extra repository is removed
        status.return_value = True
        load.side_effect = ({}, {})
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={"extra-repo": {}}),
            "pkg.del_repo": MagicMock(),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertEqual(
                freezer.restore(),
                {
                    "pkgs": {"add": [], "remove": []},
                    "repos": {"add": [], "remove": ["extra-repo"]},
                    "comment": [],
                },
            )
            salt_mock["pkg.list_pkgs"].assert_called()
            salt_mock["pkg.list_repos"].assert_called()
            salt_mock["pkg.del_repo"].assert_called_once()
            fopen.assert_called()
            load.assert_called()

    @patch("os.remove")
    @patch("salt.utils.json.load")
    @patch("salt.modules.freezer.fopen")
    @patch("salt.modules.freezer.status")
    def test_restore_clean_yml(self, status, fopen, load, remove):
        """
        Test to restore an old state
        """
        status.return_value = True
        salt_mock = {
            "pkg.list_pkgs": MagicMock(return_value={}),
            "pkg.list_repos": MagicMock(return_value={}),
            "pkg.install": MagicMock(),
        }
        with patch.dict(freezer.__salt__, salt_mock):
            self.assertEqual(
                freezer.restore(clean=True),
                {
                    "pkgs": {"add": [], "remove": []},
                    "repos": {"add": [], "remove": []},
                    "comment": [],
                },
            )
            salt_mock["pkg.list_pkgs"].assert_called()
            salt_mock["pkg.list_repos"].assert_called()
            fopen.assert_called()
            load.assert_called()
            self.assertEqual(remove.call_count, 2)



"""
Management of Linux logical volumes
===================================

A state module to manage LVMs

.. code-block:: yaml

    /dev/sda:
      lvm.pv_present

    my_vg:
      lvm.vg_present:
        - devices: /dev/sda

    lvroot:
      lvm.lv_present:
        - vgname: my_vg
        - size: 10G
        - stripes: 5
        - stripesize: 8K
"""

import os

import salt.utils.path


def __virtual__():
    """
    Only load the module if lvm is installed
    """
    if salt.utils.path.which("lvm"):
        return "lvm"
    return (False, "lvm command not found")


def _convert_to_mb(size):

    str_size = str(size).lower()
    unit = str_size[-1:]
    if unit.isdigit():
        unit = "m"
    elif unit == "b":
        unit = str_size[-2:-1]
        str_size = str_size[:-2]
    else:
        str_size = str_size[:-1]

    if str_size[-1:].isdigit():
        size = int(str_size)
    else:
        raise salt.exceptions.ArgumentValueError("Size {} is invalid.".format(size))

    if unit == "s":
        target_size = size / 2048
    elif unit == "m":
        target_size = size
    elif unit == "g":
        target_size = size * 1024
    elif unit == "t":
        target_size = size * 1024 * 1024
    elif unit == "p":
        target_size = size * 1024 * 1024 * 1024
    else:
        raise salt.exceptions.ArgumentValueError("Unit {} is invalid.".format(unit))
    return target_size


def pv_present(name, **kwargs):
    """
    Set a Physical Device to be used as an LVM Physical Volume

    name
        The device name to initialize.

    kwargs
        Any supported options to pvcreate. See
        :mod:`linux_lvm <salt.modules.linux_lvm>` for more details.
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    if __salt__["lvm.pvdisplay"](name, quiet=True):
        ret["comment"] = "Physical Volume {} already present".format(name)
    elif __opts__["test"]:
        ret["comment"] = "Physical Volume {} is set to be created".format(name)
        ret["result"] = None
        return ret
    else:
        changes = __salt__["lvm.pvcreate"](name, **kwargs)

        if __salt__["lvm.pvdisplay"](name):
            ret["comment"] = "Created Physical Volume {}".format(name)
            ret["changes"]["created"] = changes
        else:
            ret["comment"] = "Failed to create Physical Volume {}".format(name)
            ret["result"] = False
    return ret


def pv_absent(name):
    """
    Ensure that a Physical Device is not being used by lvm

    name
        The device name to initialize.
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    if not __salt__["lvm.pvdisplay"](name, quiet=True):
        ret["comment"] = "Physical Volume {} does not exist".format(name)
    elif __opts__["test"]:
        ret["comment"] = "Physical Volume {} is set to be removed".format(name)
        ret["result"] = None
        return ret
    else:
        changes = __salt__["lvm.pvremove"](name)

        if __salt__["lvm.pvdisplay"](name, quiet=True):
            ret["comment"] = "Failed to remove Physical Volume {}".format(name)
            ret["result"] = False
        else:
            ret["comment"] = "Removed Physical Volume {}".format(name)
            ret["changes"]["removed"] = changes
    return ret


def vg_present(name, devices=None, **kwargs):
    """
    Create an LVM Volume Group

    name
        The Volume Group name to create

    devices
        A list of devices that will be added to the Volume Group

    kwargs
        Any supported options to vgcreate. See
        :mod:`linux_lvm <salt.modules.linux_lvm>` for more details.
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}
    if isinstance(devices, str):
        devices = devices.split(",")

    if __salt__["lvm.vgdisplay"](name, quiet=True):
        ret["comment"] = "Volume Group {} already present".format(name)
        for device in devices:
            realdev = os.path.realpath(device)
            pvs = __salt__["lvm.pvdisplay"](realdev, real=True)
            if pvs and pvs.get(realdev, None):
                if pvs[realdev]["Volume Group Name"] == name:
                    ret["comment"] = "{}\n{}".format(
                        ret["comment"], "{} is part of Volume Group".format(device)
                    )
                elif pvs[realdev]["Volume Group Name"] in ["", "#orphans_lvm2"]:
                    __salt__["lvm.vgextend"](name, device)
                    pvs = __salt__["lvm.pvdisplay"](realdev, real=True)
                    if pvs[realdev]["Volume Group Name"] == name:
                        ret["changes"].update({device: "added to {}".format(name)})
                    else:
                        ret["comment"] = "{}\n{}".format(
                            ret["comment"], "{} could not be added".format(device)
                        )
                        ret["result"] = False
                else:
                    ret["comment"] = "{}\n{}".format(
                        ret["comment"],
                        "{} is part of {}".format(
                            device, pvs[realdev]["Volume Group Name"]
                        ),
                    )
                    ret["result"] = False
            else:
                ret["comment"] = "{}\n{}".format(
                    ret["comment"], "pv {} is not present".format(device)
                )
                ret["result"] = False
    elif __opts__["test"]:
        ret["comment"] = "Volume Group {} is set to be created".format(name)
        ret["result"] = None
        return ret
    else:
        changes = __salt__["lvm.vgcreate"](name, devices, **kwargs)

        if __salt__["lvm.vgdisplay"](name):
            ret["comment"] = "Created Volume Group {}".format(name)
            ret["changes"]["created"] = changes
        else:
            ret["comment"] = "Failed to create Volume Group {}".format(name)
            ret["result"] = False
    return ret


def vg_absent(name):
    """
    Remove an LVM volume group

    name
        The volume group to remove
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    if not __salt__["lvm.vgdisplay"](name, quiet=True):
        ret["comment"] = "Volume Group {} already absent".format(name)
    elif __opts__["test"]:
        ret["comment"] = "Volume Group {} is set to be removed".format(name)
        ret["result"] = None
        return ret
    else:
        changes = __salt__["lvm.vgremove"](name)

        if not __salt__["lvm.vgdisplay"](name, quiet=True):
            ret["comment"] = "Removed Volume Group {}".format(name)
            ret["changes"]["removed"] = changes
        else:
            ret["comment"] = "Failed to remove Volume Group {}".format(name)
            ret["result"] = False
    return ret


def lv_present(
    name,
    vgname=None,
    size=None,
    extents=None,
    snapshot=None,
    pv="",
    thinvolume=False,
    thinpool=False,
    force=False,
    resizefs=False,
    **kwargs
):
    """
    Ensure that a Logical Volume is present, creating it if absent.

    name
        The name of the Logical Volume

    vgname
        The name of the Volume Group on which the Logical Volume resides

    size
        The size of the Logical Volume in megabytes, or use a suffix
        such as S, M, G, T, P for 512 byte sectors, megabytes, gigabytes
        or terabytes respectively. The suffix is case insensitive.

    extents
        The number of logical extents allocated to the Logical Volume
        It can be a percentage allowed by lvcreate's syntax, in this case
        it will set the Logical Volume initial size and won't be resized.

    snapshot
        The name of the snapshot

    pv
        The Physical Volume to use

    kwargs
        Any supported options to lvcreate. See
        :mod:`linux_lvm <salt.modules.linux_lvm>` for more details.

    .. versionadded:: 2016.11.0

    thinvolume
        Logical Volume is thinly provisioned

    thinpool
        Logical Volume is a thin pool

    .. versionadded:: 2018.3.0

    force
        Assume yes to all prompts

    .. versionadded:: 3002.0

    resizefs
        Use fsadm to resize the logical volume filesystem if needed

    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    if extents and size:
        ret["comment"] = "Only one of extents or size can be specified."
        ret["result"] = False
        return ret

    if size:
        size_mb = _convert_to_mb(size)

    _snapshot = None

    if snapshot:
        _snapshot = name
        name = snapshot

    if thinvolume:
        lvpath = "/dev/{}/{}".format(vgname.split("/")[0], name)
    else:
        lvpath = "/dev/{}/{}".format(vgname, name)

    lv_info = __salt__["lvm.lvdisplay"](lvpath, quiet=True)
    lv_info = lv_info.get(lvpath)

    if not lv_info:
        if __opts__["test"]:
            ret["comment"] = "Logical Volume {} is set to be created".format(name)
            ret["result"] = None
            return ret
        else:
            changes = __salt__["lvm.lvcreate"](
                name,
                vgname,
                size=size,
                extents=extents,
                snapshot=_snapshot,
                pv=pv,
                thinvolume=thinvolume,
                thinpool=thinpool,
                force=force,
                **kwargs
            )

            if __salt__["lvm.lvdisplay"](lvpath):
                ret["comment"] = "Created Logical Volume {}".format(name)
                ret["changes"]["created"] = changes
            else:
                ret["comment"] = "Failed to create Logical Volume {}. Error: {}".format(
                    name, changes["Output from lvcreate"]
                )
                ret["result"] = False
    else:
        ret["comment"] = "Logical Volume {} already present".format(name)

        if size or extents:
            old_extents = int(lv_info["Current Logical Extents Associated"])
            old_size_mb = _convert_to_mb(lv_info["Logical Volume Size"] + "s")
            if size:
                extents = old_extents
            else:
                # ignore percentage "extents" if the logical volume already exists
                if "%" in str(extents):
                    ret[
                        "comment"
                    ] = "Logical Volume {} already present, {} won't be resized.".format(
                        name, extents
                    )
                    extents = old_extents
                size_mb = old_size_mb

            if force is False and (size_mb < old_size_mb or extents < old_extents):
                ret[
                    "comment"
                ] = "To reduce a Logical Volume option 'force' must be True."
                ret["result"] = False
                return ret

            if size_mb != old_size_mb or extents != old_extents:
                if __opts__["test"]:
                    ret["comment"] = "Logical Volume {} is set to be resized".format(
                        name
                    )
                    ret["result"] = None
                    return ret
                else:
                    if size:
                        changes = __salt__["lvm.lvresize"](
                            lvpath=lvpath, size=size, resizefs=resizefs, force=force
                        )
                    else:
                        changes = __salt__["lvm.lvresize"](
                            lvpath=lvpath,
                            extents=extents,
                            resizefs=resizefs,
                            force=force,
                        )

                    if not changes:
                        ret[
                            "comment"
                        ] = "Failed to resize Logical Volume. Unknown Error."
                        ret["result"] = False

                    lv_info = __salt__["lvm.lvdisplay"](lvpath, quiet=True)[lvpath]
                    new_size_mb = _convert_to_mb(lv_info["Logical Volume Size"] + "s")
                    if new_size_mb != old_size_mb:
                        ret["comment"] = "Resized Logical Volume {}".format(name)
                        ret["changes"]["resized"] = changes
                    else:
                        ret[
                            "comment"
                        ] = "Failed to resize Logical Volume {}.\nError: {}".format(
                            name, changes["Output from lvresize"]
                        )
                        ret["result"] = False
    return ret


def lv_absent(name, vgname=None):
    """
    Remove a given existing Logical Volume from a named existing volume group

    name
        The Logical Volume to remove

    vgname
        The name of the Volume Group on which the Logical Volume resides
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    lvpath = "/dev/{}/{}".format(vgname, name)
    if not __salt__["lvm.lvdisplay"](lvpath, quiet=True):
        ret["comment"] = "Logical Volume {} already absent".format(name)
    elif __opts__["test"]:
        ret["comment"] = "Logical Volume {} is set to be removed".format(name)
        ret["result"] = None
        return ret
    else:
        changes = __salt__["lvm.lvremove"](name, vgname)

        if not __salt__["lvm.lvdisplay"](lvpath, quiet=True):
            ret["comment"] = "Removed Logical Volume {}".format(name)
            ret["changes"]["removed"] = changes
        else:
            ret["comment"] = "Failed to remove Logical Volume {}".format(name)
            ret["result"] = False
    return ret



import pytest
import salt.serializers.toml


@pytest.mark.skipif(
    salt.serializers.toml.HAS_TOML is False, reason="The 'toml' library is missing"
)
def test_toml_renderer(salt_call_cli, tmp_path, base_env_state_tree_root_dir):
    config_file_path = tmp_path / "config.toml"
    state_file = """
    toml-config:
      file.serialize:
        - name: {{ pillar.get("toml-config-path") }}
        - formatter: toml
        - dataset:
            tool:
              black:
                exclude: foobar
              isort:
                include_trailing_comma: true
    """
    pillar = {
        "toml-config-path": str(config_file_path).replace("\\", "/"),
    }
    with pytest.helpers.temp_file(
        "issue-58822.sls", state_file, base_env_state_tree_root_dir
    ):
        ret = salt_call_cli.run("state.apply", "issue-58822", pillar=pillar)
        assert ret.exitcode == 0
    contents = config_file_path.read_text()
    expected = (
        '[tool.black]\nexclude = "foobar"\n\n[tool.isort]\ninclude_trailing_comma ='
        " true\n\n"
    )
    assert contents == expected



import os
import pprint

import pytest
import salt.utils.path
import salt.utils.pkg
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.helpers import requires_system_grains
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf


@pytest.mark.windows_whitelisted
class PkgModuleTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the pkg module
    """

    @classmethod
    @requires_system_grains
    def setUpClass(cls, grains):  # pylint: disable=arguments-differ
        cls.ctx = {}
        cls.pkg = "figlet"
        if salt.utils.platform.is_windows():
            cls.pkg = "putty"
        elif grains["os_family"] == "RedHat":
            if grains["os"] == "VMware Photon OS":
                cls.pkg = "snoopy"
            else:
                cls.pkg = "units"

    @pytest.mark.skip_if_not_root
    @pytest.mark.requires_salt_modules("pkg.refresh_db")
    def setUp(self):
        if "refresh" not in self.ctx:
            self.run_function("pkg.refresh_db")
            self.ctx["refresh"] = True

    @pytest.mark.requires_salt_modules("pkg.list_pkgs")
    @pytest.mark.slow_test
    def test_list(self):
        """
        verify that packages are installed
        """
        ret = self.run_function("pkg.list_pkgs")
        self.assertNotEqual(len(ret.keys()), 0)

    @pytest.mark.requires_salt_modules("pkg.version_cmp")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_version_cmp(self, grains):
        """
        test package version comparison on supported platforms
        """
        func = "pkg.version_cmp"
        if grains["os_family"] == "Debian":
            lt = ["0.2.4-0ubuntu1", "0.2.4.1-0ubuntu1"]
            eq = ["0.2.4-0ubuntu1", "0.2.4-0ubuntu1"]
            gt = ["0.2.4.1-0ubuntu1", "0.2.4-0ubuntu1"]
        elif grains["os_family"] == "Suse":
            lt = ["2.3.0-1", "2.3.1-15.1"]
            eq = ["2.3.1-15.1", "2.3.1-15.1"]
            gt = ["2.3.2-15.1", "2.3.1-15.1"]
        else:
            lt = ["2.3.0", "2.3.1"]
            eq = ["2.3.1", "2.3.1"]
            gt = ["2.3.2", "2.3.1"]

        self.assertEqual(self.run_function(func, lt), -1)
        self.assertEqual(self.run_function(func, eq), 0)
        self.assertEqual(self.run_function(func, gt), 1)

    @pytest.mark.destructive_test
    @pytest.mark.requires_salt_modules("pkg.mod_repo", "pkg.del_repo", "pkg.get_repo")
    @requires_system_grains
    @pytest.mark.slow_test
    @pytest.mark.requires_network
    def test_mod_del_repo(self, grains):
        """
        test modifying and deleting a software repository
        """
        repo = None

        try:
            if grains["os"] == "Ubuntu":
                repo = "ppa:otto-kesselgulasch/gimp-edge"
                uri = "http://ppa.launchpad.net/otto-kesselgulasch/gimp-edge/ubuntu"
                ret = self.run_function("pkg.mod_repo", [repo, "comps=main"])
                self.assertNotEqual(ret, {})
                ret = self.run_function("pkg.get_repo", [repo])

                self.assertIsInstance(
                    ret,
                    dict,
                    "The 'pkg.get_repo' command did not return the excepted dictionary."
                    " Output:\n{}".format(ret),
                )
                self.assertEqual(
                    ret["uri"],
                    uri,
                    msg="The URI did not match. Full return:\n{}".format(
                        pprint.pformat(ret)
                    ),
                )
            elif grains["os_family"] == "RedHat":
                repo = "saltstack"
                name = "SaltStack repo for RHEL/CentOS {}".format(
                    grains["osmajorrelease"]
                )
                baseurl = (
                    "http://repo.saltproject.io/py3/redhat/{}/x86_64/latest/".format(
                        grains["osmajorrelease"]
                    )
                )
                gpgkey = "https://repo.saltproject.io/py3/redhat/{}/x86_64/latest/SALTSTACK-GPG-KEY.pub".format(
                    grains["osmajorrelease"]
                )
                gpgcheck = 1
                enabled = 1
                ret = self.run_function(
                    "pkg.mod_repo",
                    [repo],
                    name=name,
                    baseurl=baseurl,
                    gpgkey=gpgkey,
                    gpgcheck=gpgcheck,
                    enabled=enabled,
                )
                # return data from pkg.mod_repo contains the file modified at
                # the top level, so use next(iter(ret)) to get that key
                self.assertNotEqual(ret, {})
                repo_info = ret[next(iter(ret))]
                self.assertIn(repo, repo_info)
                self.assertEqual(repo_info[repo]["baseurl"], baseurl)
                ret = self.run_function("pkg.get_repo", [repo])
                self.assertEqual(ret["baseurl"], baseurl)
        finally:
            if repo is not None:
                self.run_function("pkg.del_repo", [repo])

    @pytest.mark.slow_test
    def test_mod_del_repo_multiline_values(self):
        """
        test modifying and deleting a software repository defined with multiline values
        """
        os_grain = self.run_function("grains.item", ["os"])["os"]
        repo = None
        try:
            if os_grain in ["CentOS", "RedHat", "VMware Photon OS"]:
                my_baseurl = (
                    "http://my.fake.repo/foo/bar/\n http://my.fake.repo.alt/foo/bar/"
                )
                expected_get_repo_baseurl = (
                    "http://my.fake.repo/foo/bar/\nhttp://my.fake.repo.alt/foo/bar/"
                )
                major_release = int(
                    self.run_function("grains.item", ["osmajorrelease"])[
                        "osmajorrelease"
                    ]
                )
                repo = "fakerepo"
                name = "Fake repo for RHEL/CentOS/SUSE"
                baseurl = my_baseurl
                gpgkey = "https://my.fake.repo/foo/bar/MY-GPG-KEY.pub"
                failovermethod = "priority"
                gpgcheck = 1
                enabled = 1
                ret = self.run_function(
                    "pkg.mod_repo",
                    [repo],
                    name=name,
                    baseurl=baseurl,
                    gpgkey=gpgkey,
                    gpgcheck=gpgcheck,
                    enabled=enabled,
                    failovermethod=failovermethod,
                )
                # return data from pkg.mod_repo contains the file modified at
                # the top level, so use next(iter(ret)) to get that key
                self.assertNotEqual(ret, {})
                repo_info = ret[next(iter(ret))]
                self.assertIn(repo, repo_info)
                self.assertEqual(repo_info[repo]["baseurl"], my_baseurl)
                ret = self.run_function("pkg.get_repo", [repo])
                self.assertEqual(ret["baseurl"], expected_get_repo_baseurl)
                self.run_function("pkg.mod_repo", [repo])
                ret = self.run_function("pkg.get_repo", [repo])
                self.assertEqual(ret["baseurl"], expected_get_repo_baseurl)
        finally:
            if repo is not None:
                self.run_function("pkg.del_repo", [repo])

    @pytest.mark.requires_salt_modules("pkg.owner")
    def test_owner(self):
        """
        test finding the package owning a file
        """
        func = "pkg.owner"
        ret = self.run_function(func, ["/bin/ls"])
        self.assertNotEqual(len(ret), 0)

    # Similar to pkg.owner, but for FreeBSD's pkgng
    @pytest.mark.requires_salt_modules("pkg.which")
    def test_which(self):
        """
        test finding the package owning a file
        """
        func = "pkg.which"
        ret = self.run_function(func, ["/usr/local/bin/salt-call"])
        self.assertNotEqual(len(ret), 0)

    @pytest.mark.destructive_test
    @pytest.mark.requires_salt_modules("pkg.version", "pkg.install", "pkg.remove")
    @pytest.mark.slow_test
    @pytest.mark.requires_network
    def test_install_remove(self):
        """
        successfully install and uninstall a package
        """
        version = self.run_function("pkg.version", [self.pkg])

        def test_install():
            install_ret = self.run_function("pkg.install", [self.pkg])
            self.assertIn(self.pkg, install_ret)

        def test_remove():
            remove_ret = self.run_function("pkg.remove", [self.pkg])
            self.assertIn(self.pkg, remove_ret)

        if version and isinstance(version, dict):
            version = version[self.pkg]

        if version:
            test_remove()
            test_install()
        else:
            test_install()
            test_remove()

    @pytest.mark.destructive_test
    @pytest.mark.skipif(
        salt.utils.platform.is_photonos(),
        reason="package hold/unhold unsupported on Photon OS",
    )
    @pytest.mark.requires_salt_modules(
        "pkg.hold",
        "pkg.unhold",
        "pkg.install",
        "pkg.version",
        "pkg.remove",
        "pkg.list_pkgs",
    )
    @requires_system_grains
    @pytest.mark.slow_test
    @pytest.mark.requires_network
    @pytest.mark.requires_salt_states("pkg.installed")
    def test_hold_unhold(self, grains):
        """
        test holding and unholding a package
        """
        versionlock_pkg = None
        if grains["os_family"] == "RedHat":
            pkgs = {
                p for p in self.run_function("pkg.list_pkgs") if "-versionlock" in p
            }
            if not pkgs:
                self.skipTest("No versionlock package found in repositories")
            for versionlock_pkg in pkgs:
                ret = self.run_state(
                    "pkg.installed", name=versionlock_pkg, refresh=False
                )
                # Exit loop if a versionlock package installed correctly
                try:
                    self.assertSaltTrueReturn(ret)
                    break
                except AssertionError:
                    pass
            else:
                self.fail("Could not install versionlock package from {}".format(pkgs))

        self.run_function("pkg.install", [self.pkg])

        try:
            hold_ret = self.run_function("pkg.hold", [self.pkg])
            if versionlock_pkg and "-versionlock is not installed" in str(hold_ret):
                self.skipTest("{}  `{}` is installed".format(hold_ret, versionlock_pkg))
            self.assertIn(self.pkg, hold_ret)
            self.assertTrue(hold_ret[self.pkg]["result"])

            unhold_ret = self.run_function("pkg.unhold", [self.pkg])
            self.assertIn(self.pkg, unhold_ret)
            self.assertTrue(unhold_ret[self.pkg]["result"])
            self.run_function("pkg.remove", [self.pkg])
        finally:
            if versionlock_pkg:
                ret = self.run_state("pkg.removed", name=versionlock_pkg)
                self.assertSaltTrueReturn(ret)

    @pytest.mark.destructive_test
    @pytest.mark.requires_salt_modules("pkg.refresh_db")
    @requires_system_grains
    @pytest.mark.slow_test
    @pytest.mark.requires_network
    def test_refresh_db(self, grains):
        """
        test refreshing the package database
        """
        func = "pkg.refresh_db"

        rtag = salt.utils.pkg.rtag(self.minion_opts)
        salt.utils.pkg.write_rtag(self.minion_opts)
        self.assertTrue(os.path.isfile(rtag))

        ret = self.run_function(func)
        if not isinstance(ret, dict):
            self.skipTest(
                "Upstream repo did not return coherent results: {}".format(ret)
            )

        if grains["os_family"] == "RedHat":
            self.assertIn(ret, (True, None))
        elif grains["os_family"] == "Suse":
            if not isinstance(ret, dict):
                self.skipTest(
                    "Upstream repo did not return coherent results. Skipping test."
                )
            self.assertNotEqual(ret, {})
            for source, state in ret.items():
                self.assertIn(state, (True, False, None))

        self.assertFalse(os.path.isfile(rtag))

    @pytest.mark.requires_salt_modules("pkg.info_installed")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_pkg_info(self, grains):
        """
        Test returning useful information on Ubuntu systems.
        """
        func = "pkg.info_installed"

        if grains["os_family"] == "Debian":
            ret = self.run_function(func, ["bash", "dpkg"])
            keys = ret.keys()
            self.assertIn("bash", keys)
            self.assertIn("dpkg", keys)
        elif grains["os_family"] == "RedHat":
            ret = self.run_function(func, ["rpm", "bash"])
            keys = ret.keys()
            self.assertIn("rpm", keys)
            self.assertIn("bash", keys)
        elif grains["os_family"] == "Suse":
            ret = self.run_function(func, ["less", "zypper"])
            keys = ret.keys()
            self.assertIn("less", keys)
            self.assertIn("zypper", keys)
        else:
            ret = self.run_function(func, [self.pkg])
            keys = ret.keys()
            self.assertIn(self.pkg, keys)

    @skipIf(True, "Temporary Skip - Causes centos 8 test to fail")
    @pytest.mark.destructive_test
    @pytest.mark.requires_salt_modules(
        "pkg.refresh_db",
        "pkg.upgrade",
        "pkg.install",
        "pkg.list_repo_pkgs",
        "pkg.list_upgrades",
    )
    @requires_system_grains
    @pytest.mark.slow_test
    @pytest.mark.requires_network
    def test_pkg_upgrade_has_pending_upgrades(self, grains):
        """
        Test running a system upgrade when there are packages that need upgrading
        """
        if grains["os"] == "Arch":
            self.skipTest("Arch moved to Python 3.8 and we're not ready for it yet")

        func = "pkg.upgrade"

        # First make sure that an up-to-date copy of the package db is available
        self.run_function("pkg.refresh_db")

        if grains["os_family"] == "Suse":
            # This test assumes that there are multiple possible versions of a
            # package available. That makes it brittle if you pick just one
            # target, as changes in the available packages will break the test.
            # Therefore, we'll choose from several packages to make sure we get
            # one that is suitable for this test.
            packages = ("hwinfo", "avrdude", "diffoscope", "vim")
            available = self.run_function("pkg.list_repo_pkgs", packages)

            for package in packages:
                try:
                    new, old = available[package][:2]
                except (KeyError, ValueError):
                    # Package not available, or less than 2 versions
                    # available. This is not a suitable target.
                    continue
                else:
                    target = package
                    break
            else:
                # None of the packages have more than one version available, so
                # we need to find new package(s). pkg.list_repo_pkgs can be
                # used to get an overview of the available packages. We should
                # try to find packages with few dependencies and small download
                # sizes, to keep this test from taking longer than necessary.
                self.fail("No suitable package found for this test")

            # Make sure we have the 2nd-oldest available version installed
            ret = self.run_function("pkg.install", [target], version=old)
            if not isinstance(ret, dict):
                if ret.startswith("ERROR"):
                    self.skipTest(
                        "Could not install older {} to complete test.".format(target)
                    )

            # Run a system upgrade, which should catch the fact that the
            # targeted package needs upgrading, and upgrade it.
            ret = self.run_function(func)

            # The changes dictionary should not be empty.
            if "changes" in ret:
                self.assertIn(target, ret["changes"])
            else:
                self.assertIn(target, ret)
        else:
            ret = self.run_function("pkg.list_upgrades")
            if ret == "" or ret == {}:
                self.skipTest(
                    "No updates available for this machine.  Skipping pkg.upgrade test."
                )
            else:
                args = []
                if grains["os_family"] == "Debian":
                    args = ["dist_upgrade=True"]
                ret = self.run_function(func, args)
                self.assertNotEqual(ret, {})

    @pytest.mark.destructive_test
    @skipIf(
        salt.utils.platform.is_darwin(),
        "The jenkins user is equivalent to root on mac, causing the test to be"
        " unrunnable",
    )
    @pytest.mark.requires_salt_modules("pkg.remove", "pkg.latest_version")
    @requires_system_grains
    @pytest.mark.slow_test
    @pytest.mark.requires_salt_states("pkg.removed")
    def test_pkg_latest_version(self, grains):
        """
        Check that pkg.latest_version returns the latest version of the uninstalled package.
        The package is not installed. Only the package version is checked.
        """
        self.run_state("pkg.removed", name=self.pkg)

        cmd_pkg = []
        if grains["os_family"] == "RedHat":
            cmd_pkg = self.run_function("cmd.run", ["yum list {}".format(self.pkg)])
        elif salt.utils.platform.is_windows():
            cmd_pkg = self.run_function("pkg.list_available", [self.pkg])
        elif grains["os_family"] == "Debian":
            cmd_pkg = self.run_function("cmd.run", ["apt list {}".format(self.pkg)])
        elif grains["os_family"] == "Arch":
            cmd_pkg = self.run_function("cmd.run", ["pacman -Si {}".format(self.pkg)])
        elif grains["os_family"] == "FreeBSD":
            cmd_pkg = self.run_function(
                "cmd.run", ["pkg search -S name -qQ version -e {}".format(self.pkg)]
            )
        elif grains["os_family"] == "Suse":
            cmd_pkg = self.run_function("cmd.run", ["zypper info {}".format(self.pkg)])
        elif grains["os_family"] == "MacOS":
            brew_bin = salt.utils.path.which("brew")
            mac_user = self.run_function("file.get_user", [brew_bin])
            if mac_user == "root":
                self.skipTest(
                    "brew cannot run as root, try a user in {}".format(
                        os.listdir("/Users/")
                    )
                )
            cmd_pkg = self.run_function(
                "cmd.run", ["brew info {}".format(self.pkg)], run_as=mac_user
            )
        else:
            self.skipTest(
                "TODO: test not configured for {}".format(grains["os_family"])
            )
        pkg_latest = self.run_function("pkg.latest_version", [self.pkg])
        self.assertIn(pkg_latest, cmd_pkg)



"""
Dimension Data Cloud Module
===========================

This is a cloud module for the Dimension Data Cloud,
using the existing Libcloud driver for Dimension Data.

.. code-block:: yaml

    # Note: This example is for /etc/salt/cloud.providers
    # or any file in the
    # /etc/salt/cloud.providers.d/ directory.

    my-dimensiondata-config:
      user_id: my_username
      key: myPassword!
      region: dd-na
      driver: dimensiondata

:maintainer: Anthony Shaw <anthonyshaw@apache.org>
:depends: libcloud >= 1.2.1
"""

import logging
import pprint
import socket

import salt.config as config
import salt.utils.cloud
from salt.cloud.libcloudfuncs import *  # pylint: disable=redefined-builtin,wildcard-import,unused-wildcard-import
from salt.exceptions import (
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudSystemExit,
)
from salt.utils.functools import namespaced_function
from salt.utils.versions import LooseVersion as _LooseVersion

# Import libcloud
try:
    import libcloud
    from libcloud.compute.base import NodeDriver, NodeState
    from libcloud.compute.base import NodeAuthPassword
    from libcloud.compute.types import Provider
    from libcloud.compute.providers import get_driver
    from libcloud.loadbalancer.base import Member
    from libcloud.loadbalancer.types import Provider as Provider_lb
    from libcloud.loadbalancer.providers import get_driver as get_driver_lb

    # This work-around for Issue #32743 is no longer needed for libcloud >=
    # 1.4.0. However, older versions of libcloud must still be supported with
    # this work-around. This work-around can be removed when the required
    # minimum version of libcloud is 2.0.0 (See PR #40837 - which is
    # implemented in Salt 2018.3.0).
    if _LooseVersion(libcloud.__version__) < _LooseVersion("1.4.0"):
        # See https://github.com/saltstack/salt/issues/32743
        import libcloud.security

        libcloud.security.CA_CERTS_PATH.append("/etc/ssl/certs/YaST-CA.pem")
    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False


try:
    from netaddr import all_matching_cidrs  # pylint: disable=unused-import

    HAS_NETADDR = True
except ImportError:
    HAS_NETADDR = False


# Some of the libcloud functions need to be in the same namespace as the
# functions defined in the module, so we create new function objects inside
# this module namespace
get_size = namespaced_function(get_size, globals())
get_image = namespaced_function(get_image, globals())
avail_locations = namespaced_function(avail_locations, globals())
avail_images = namespaced_function(avail_images, globals())
avail_sizes = namespaced_function(avail_sizes, globals())
script = namespaced_function(script, globals())
destroy = namespaced_function(destroy, globals())
reboot = namespaced_function(reboot, globals())
list_nodes = namespaced_function(list_nodes, globals())
list_nodes_full = namespaced_function(list_nodes_full, globals())
list_nodes_select = namespaced_function(list_nodes_select, globals())
show_instance = namespaced_function(show_instance, globals())
get_node = namespaced_function(get_node, globals())

# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "dimensiondata"


def __virtual__():
    """
    Set up the libcloud functions and check for dimensiondata configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    for provider, details in __opts__["providers"].items():
        if "dimensiondata" not in details:
            continue

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        _get_active_provider_name() or "dimensiondata",
        ("user_id", "key", "region"),
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    deps = {"libcloud": HAS_LIBCLOUD, "netaddr": HAS_NETADDR}
    return config.check_driver_dependencies(__virtualname__, deps)


def _query_node_data(vm_, data):
    running = False
    try:
        node = show_instance(vm_["name"], "action")  # pylint: disable=not-callable
        running = node["state"] == NodeState.RUNNING
        log.debug(
            "Loaded node data for %s:\nname: %s\nstate: %s",
            vm_["name"],
            pprint.pformat(node["name"]),
            node["state"],
        )
    except Exception as err:  # pylint: disable=broad-except
        log.error(
            "Failed to get nodes list: %s",
            err,
            # Show the traceback if the debug logging level is enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        # Trigger a failure in the wait for IP function
        return running

    if not running:
        # Still not running, trigger another iteration
        return

    private = node["private_ips"]
    public = node["public_ips"]

    if private and not public:
        log.warning(
            "Private IPs returned, but not public. Checking for misidentified IPs."
        )
        for private_ip in private:
            private_ip = preferred_ip(vm_, [private_ip])
            if private_ip is False:
                continue
            if salt.utils.cloud.is_public_ip(private_ip):
                log.warning("%s is a public IP", private_ip)
                data.public_ips.append(private_ip)
            else:
                log.warning("%s is a private IP", private_ip)
                if private_ip not in data.private_ips:
                    data.private_ips.append(private_ip)

        if ssh_interface(vm_) == "private_ips" and data.private_ips:
            return data

    if private:
        data.private_ips = private
        if ssh_interface(vm_) == "private_ips":
            return data

    if public:
        data.public_ips = public
        if ssh_interface(vm_) != "private_ips":
            return data

    log.debug("Contents of the node data:")
    log.debug(data)


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__, _get_active_provider_name() or "dimensiondata", vm_["profile"]
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.info("Creating Cloud VM %s", vm_["name"])
    conn = get_conn()

    location = conn.ex_get_location_by_id(vm_["location"])
    images = conn.list_images(location=location)
    image = [x for x in images if x.id == vm_["image"]][0]
    network_domains = conn.ex_list_network_domains(location=location)
    try:
        network_domain = [
            y for y in network_domains if y.name == vm_["network_domain"]
        ][0]
    except IndexError:
        network_domain = conn.ex_create_network_domain(
            location=location,
            name=vm_["network_domain"],
            plan="ADVANCED",
            description="",
        )

    try:
        vlan = [
            y
            for y in conn.ex_list_vlans(
                location=location, network_domain=network_domain
            )
            if y.name == vm_["vlan"]
        ][0]
    except (IndexError, KeyError):
        # Use the first VLAN in the network domain
        vlan = conn.ex_list_vlans(location=location, network_domain=network_domain)[0]

    kwargs = {
        "name": vm_["name"],
        "image": image,
        "ex_description": vm_["description"],
        "ex_network_domain": network_domain,
        "ex_vlan": vlan,
        "ex_is_started": vm_["is_started"],
    }

    event_data = _to_event_data(kwargs)

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "requesting", event_data, list(event_data)
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    # Initial password (excluded from event payload)
    initial_password = NodeAuthPassword(vm_["auth"])
    kwargs["auth"] = initial_password

    try:
        data = conn.create_node(**kwargs)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error creating %s on DIMENSIONDATA\n\n"
            "The following exception was thrown by libcloud when trying to "
            "run the initial deployment: \n%s",
            vm_["name"],
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False

    try:
        data = __utils__["cloud.wait_for_ip"](
            _query_node_data,
            update_args=(vm_, data),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=25 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=30
            ),
            max_failures=config.get_cloud_config_value(
                "wait_for_ip_max_failures", vm_, __opts__, default=60
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])  # pylint: disable=not-callable
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc))

    log.debug("VM is now running")
    if ssh_interface(vm_) == "private_ips":
        ip_address = preferred_ip(vm_, data.private_ips)
    else:
        ip_address = preferred_ip(vm_, data.public_ips)
    log.debug("Using IP address %s", ip_address)

    if __utils__["cloud.get_salt_interface"](vm_, __opts__) == "private_ips":
        salt_ip_address = preferred_ip(vm_, data.private_ips)
        log.info("Salt interface set to: %s", salt_ip_address)
    else:
        salt_ip_address = preferred_ip(vm_, data.public_ips)
        log.debug("Salt interface set to: %s", salt_ip_address)

    if not ip_address:
        raise SaltCloudSystemExit("No IP addresses could be found.")

    vm_["salt_host"] = salt_ip_address
    vm_["ssh_host"] = ip_address
    vm_["password"] = vm_["auth"]

    ret = __utils__["cloud.bootstrap"](vm_, __opts__)

    ret.update(data.__dict__)

    if "password" in data.extra:
        del data.extra["password"]

    log.info("Created Cloud VM '%s'", vm_["name"])
    log.debug(
        "'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data.__dict__)
    )

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def create_lb(kwargs=None, call=None):
    r"""
    Create a load-balancer configuration.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f create_lb dimensiondata \
            name=dev-lb port=80 protocol=http \
            members=w1,w2,w3 algorithm=ROUND_ROBIN
    """
    conn = get_conn()
    if call != "function":
        raise SaltCloudSystemExit(
            "The create_lb function must be called with -f or --function."
        )

    if not kwargs or "name" not in kwargs:
        log.error("A name must be specified when creating a health check.")
        return False
    if "port" not in kwargs:
        log.error("A port or port-range must be specified for the load-balancer.")
        return False
    if "networkdomain" not in kwargs:
        log.error("A network domain must be specified for the load-balancer.")
        return False
    if "members" in kwargs:
        members = []
        ip = ""
        membersList = kwargs.get("members").split(",")
        log.debug("MemberList: %s", membersList)
        for member in membersList:
            try:
                log.debug("Member: %s", member)
                node = get_node(conn, member)  # pylint: disable=not-callable
                log.debug("Node: %s", node)
                ip = node.private_ips[0]
            except Exception as err:  # pylint: disable=broad-except
                log.error(
                    "Failed to get node ip: %s",
                    err,
                    # Show the traceback if the debug logging level is enabled
                    exc_info_on_loglevel=logging.DEBUG,
                )
            members.append(Member(ip, ip, kwargs["port"]))
    else:
        members = None
    log.debug("Members: %s", members)

    networkdomain = kwargs["networkdomain"]
    name = kwargs["name"]
    port = kwargs["port"]
    protocol = kwargs.get("protocol", None)
    algorithm = kwargs.get("algorithm", None)

    lb_conn = get_lb_conn(conn)
    network_domains = conn.ex_list_network_domains()
    network_domain = [y for y in network_domains if y.name == networkdomain][0]

    log.debug("Network Domain: %s", network_domain.id)
    lb_conn.ex_set_current_network_domain(network_domain.id)

    event_data = _to_event_data(kwargs)

    __utils__["cloud.fire_event"](
        "event",
        "create load_balancer",
        "salt/cloud/loadbalancer/creating",
        args=event_data,
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    lb = lb_conn.create_balancer(name, port, protocol, algorithm, members)

    event_data = _to_event_data(kwargs)

    __utils__["cloud.fire_event"](
        "event",
        "created load_balancer",
        "salt/cloud/loadbalancer/created",
        args=event_data,
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )
    return _expand_balancer(lb)


def _expand_balancer(lb):
    """
    Convert the libcloud load-balancer object into something more serializable.
    """
    ret = {}
    ret.update(lb.__dict__)
    return ret


def preferred_ip(vm_, ips):
    """
    Return the preferred Internet protocol. Either 'ipv4' (default) or 'ipv6'.
    """
    proto = config.get_cloud_config_value(
        "protocol", vm_, __opts__, default="ipv4", search_global=False
    )
    family = socket.AF_INET
    if proto == "ipv6":
        family = socket.AF_INET6
    for ip in ips:
        try:
            socket.inet_pton(family, ip)
            return ip
        except Exception:  # pylint: disable=broad-except
            continue
    return False


def ssh_interface(vm_):
    """
    Return the ssh_interface type to connect to. Either 'public_ips' (default)
    or 'private_ips'.
    """
    return config.get_cloud_config_value(
        "ssh_interface", vm_, __opts__, default="public_ips", search_global=False
    )


def stop(name, call=None):
    """
    Stop a VM in DimensionData.

    name:
        The name of the VM to stop.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop vm_name
    """
    conn = get_conn()
    node = get_node(conn, name)  # pylint: disable=not-callable
    log.debug("Node of Cloud VM: %s", node)

    status = conn.ex_shutdown_graceful(node)
    log.debug("Status of Cloud VM: %s", status)

    return status


def start(name, call=None):
    """
    Stop a VM in DimensionData.

    :param str name:
        The name of the VM to stop.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop vm_name
    """

    conn = get_conn()
    node = get_node(conn, name)  # pylint: disable=not-callable
    log.debug("Node of Cloud VM: %s", node)

    status = conn.ex_start_node(node)
    log.debug("Status of Cloud VM: %s", status)

    return status


def get_conn():
    """
    Return a conn object for the passed VM data
    """
    vm_ = get_configured_provider()
    driver = get_driver(Provider.DIMENSIONDATA)

    region = config.get_cloud_config_value("region", vm_, __opts__)

    user_id = config.get_cloud_config_value("user_id", vm_, __opts__)
    key = config.get_cloud_config_value("key", vm_, __opts__)

    if key is not None:
        log.debug("DimensionData authenticating using password")

    return driver(user_id, key, region=region)


def get_lb_conn(dd_driver=None):
    """
    Return a load-balancer conn object
    """
    vm_ = get_configured_provider()

    region = config.get_cloud_config_value("region", vm_, __opts__)

    user_id = config.get_cloud_config_value("user_id", vm_, __opts__)
    key = config.get_cloud_config_value("key", vm_, __opts__)
    if not dd_driver:
        raise SaltCloudSystemExit(
            "Missing dimensiondata_driver for get_lb_conn method."
        )
    return get_driver_lb(Provider_lb.DIMENSIONDATA)(user_id, key, region=region)


def _to_event_data(obj):
    """
    Convert the specified object into a form that can be serialised by msgpack as event data.

    :param obj: The object to convert.
    """

    if obj is None:
        return None
    if isinstance(obj, bool):
        return obj
    if isinstance(obj, int):
        return obj
    if isinstance(obj, float):
        return obj
    if isinstance(obj, str):
        return obj
    if isinstance(obj, bytes):
        return obj
    if isinstance(obj, dict):
        return obj

    if isinstance(obj, NodeDriver):  # Special case for NodeDriver (cyclic references)
        return obj.name

    if isinstance(obj, list):
        return [_to_event_data(item) for item in obj]

    event_data = {}
    for attribute_name in dir(obj):
        if attribute_name.startswith("_"):
            continue

        attribute_value = getattr(obj, attribute_name)

        if callable(attribute_value):  # Strip out methods
            continue

        event_data[attribute_name] = _to_event_data(attribute_value)

    return event_data



"""
unittests for yaml outputter
"""


import salt.output.yaml_out as yaml
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class YamlTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.output.json_out
    """

    def setup_loader_modules(self):
        return {yaml: {}}

    def setUp(self):
        self.data = {"test": "two", "example": "one"}
        self.addCleanup(delattr, self, "data")

    def test_default_output(self):
        ret = yaml.output(self.data)
        expect = "example: one\ntest: two\n"
        self.assertEqual(expect, ret)

    def test_negative_int_output(self):
        with patch.dict(yaml.__opts__, {"output_indent": -1}):
            ret = yaml.output(self.data)
            expect = "{example: one, test: two}\n"
            self.assertEqual(expect, ret)



"""
Test cases for keystore state
"""

import salt.states.keystore as keystore
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class KeystoreTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.keystore
    """

    def setup_loader_modules(self):
        return {keystore: {"__opts__": {"test": False}}}

    @patch("os.path.exists", MagicMock(return_value=True))
    def test_cert_already_present(self):
        """
        Test for existing value_present
        """

        cert_return = [
            {
                "valid_until": "August 21 2017",
                "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5",
                "valid_start": "August 22 2012",
                "type": "TrustedCertEntry",
                "alias": "stringhost",
                "expired": True,
            }
        ]
        x509_return = {
            "Not After": "2017-08-21 05:26:54",
            "Subject Hash": "97:95:14:4F",
            "Serial Number": "0D:FA",
            "SHA1 Finger Print": (
                "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5"
            ),
            "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7",
            "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2",
            "Version": 1,
            "Key Size": 512,
            "Public Key": (
                "-----BEGIN PUBLIC"
                " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
                " PUBLIC KEY-----\n"
            ),
            "Issuer": {
                "C": "JP",
                "organizationName": "Frank4DD",
                "CN": "Frank4DD Web CA",
                "SP": "Tokyo",
                "L": "Chuo-ku",
                "emailAddress": "support@frank4dd.com",
                "OU": "WebCert Support",
            },
            "Issuer Hash": "92:DA:45:6B",
            "Not Before": "2012-08-22 05:26:54",
            "Subject": {
                "C": "JP",
                "SP": "Tokyo",
                "organizationName": "Frank4DD",
                "CN": "www.example.com",
            },
        }

        name = "keystore.jks"
        passphrase = "changeit"
        entries = [
            {
                "alias": "stringhost",
                "certificate": """-----BEGIN CERTIFICATE-----
                   MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
                   A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
                   MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
                   YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
                   ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
                   CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
                   ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
                   8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
                   AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
                   8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
                   2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
                   Hn+GmxZA\n-----END CERTIFICATE-----""",
            }
        ]

        state_return = {
            "name": name,
            "changes": {},
            "result": True,
            "comment": "No changes made.\n",
        }

        # with patch.dict(keystore.__opts__, {'test': False}):
        with patch.dict(
            keystore.__salt__, {"keystore.list": MagicMock(return_value=cert_return)}
        ):
            with patch.dict(
                keystore.__salt__,
                {"x509.read_certificate": MagicMock(return_value=x509_return)},
            ):
                self.assertDictEqual(
                    keystore.managed(name, passphrase, entries), state_return
                )

        with patch.dict(keystore.__opts__, {"test": True}):
            with patch.dict(
                keystore.__salt__,
                {"keystore.list": MagicMock(return_value=cert_return)},
            ):
                with patch.dict(
                    keystore.__salt__,
                    {"x509.read_certificate": MagicMock(return_value=x509_return)},
                ):
                    self.assertDictEqual(
                        keystore.managed(name, passphrase, entries), state_return
                    )

    @patch("os.path.exists", MagicMock(return_value=True))
    def test_cert_update(self):
        """
        Test for existing value_present
        """

        cert_return = [
            {
                "valid_until": "August 21 2017",
                "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5",
                "valid_start": "August 22 2012",
                "type": "TrustedCertEntry",
                "alias": "stringhost",
                "expired": True,
            }
        ]
        x509_return = {
            "Not After": "2017-08-21 05:26:54",
            "Subject Hash": "97:95:14:4F",
            "Serial Number": "0D:FA",
            "SHA1 Finger Print": (
                "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D6"
            ),
            "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7",
            "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2",
            "Version": 1,
            "Key Size": 512,
            "Public Key": (
                "-----BEGIN PUBLIC"
                " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
                " PUBLIC KEY-----\n"
            ),
            "Issuer": {
                "C": "JP",
                "organizationName": "Frank4DD",
                "CN": "Frank4DD Web CA",
                "SP": "Tokyo",
                "L": "Chuo-ku",
                "emailAddress": "support@frank4dd.com",
                "OU": "WebCert Support",
            },
            "Issuer Hash": "92:DA:45:6B",
            "Not Before": "2012-08-22 05:26:54",
            "Subject": {
                "C": "JP",
                "SP": "Tokyo",
                "organizationName": "Frank4DD",
                "CN": "www.example.com",
            },
        }

        name = "keystore.jks"
        passphrase = "changeit"
        entries = [
            {
                "alias": "stringhost",
                "certificate": """-----BEGIN CERTIFICATE-----
                   MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
                   A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
                   MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
                   YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
                   ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
                   CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
                   ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
                   8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
                   AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
                   8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
                   2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
                   Hn+GmxZA\n-----END CERTIFICATE-----""",
            }
        ]

        test_return = {
            "name": name,
            "changes": {},
            "result": None,
            "comment": "Alias stringhost would have been updated\n",
        }
        state_return = {
            "name": name,
            "changes": {"stringhost": "Updated"},
            "result": True,
            "comment": "Alias stringhost updated.\n",
        }

        with patch.dict(keystore.__opts__, {"test": True}):
            with patch.dict(
                keystore.__salt__,
                {"keystore.list": MagicMock(return_value=cert_return)},
            ):
                with patch.dict(
                    keystore.__salt__,
                    {"x509.read_certificate": MagicMock(return_value=x509_return)},
                ):
                    self.assertDictEqual(
                        keystore.managed(name, passphrase, entries), test_return
                    )

        with patch.dict(
            keystore.__salt__, {"keystore.list": MagicMock(return_value=cert_return)}
        ):
            with patch.dict(
                keystore.__salt__,
                {"x509.read_certificate": MagicMock(return_value=x509_return)},
            ):
                with patch.dict(
                    keystore.__salt__, {"keystore.remove": MagicMock(return_value=True)}
                ):
                    with patch.dict(
                        keystore.__salt__,
                        {"keystore.add": MagicMock(return_value=True)},
                    ):
                        self.assertDictEqual(
                            keystore.managed(name, passphrase, entries), state_return
                        )

    @patch("os.path.exists", MagicMock(return_value=False))
    def test_new_file(self):
        """
        Test for existing value_present
        """
        name = "keystore.jks"
        passphrase = "changeit"
        entries = [
            {
                "alias": "stringhost",
                "certificate": """-----BEGIN CERTIFICATE-----
                   MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
                   A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
                   MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
                   YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
                   ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
                   CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
                   ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
                   8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
                   AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
                   8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
                   2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
                   Hn+GmxZA\n-----END CERTIFICATE-----""",
            }
        ]

        test_return = {
            "name": name,
            "changes": {},
            "result": None,
            "comment": "Alias stringhost would have been added\n",
        }
        state_return = {
            "name": name,
            "changes": {"stringhost": "Added"},
            "result": True,
            "comment": "Alias stringhost added.\n",
        }

        with patch.dict(keystore.__opts__, {"test": True}):
            self.assertDictEqual(
                keystore.managed(name, passphrase, entries), test_return
            )

        with patch.dict(
            keystore.__salt__, {"keystore.remove": MagicMock(return_value=True)}
        ):
            with patch.dict(
                keystore.__salt__, {"keystore.add": MagicMock(return_value=True)}
            ):
                self.assertDictEqual(
                    keystore.managed(name, passphrase, entries), state_return
                )

    @patch("os.path.exists", MagicMock(return_value=True))
    def test_force_remove(self):
        """
        Test for existing value_present
        """

        cert_return = [
            {
                "valid_until": "August 21 2017",
                "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5",
                "valid_start": "August 22 2012",
                "type": "TrustedCertEntry",
                "alias": "oldhost",
                "expired": True,
            }
        ]
        x509_return = {
            "Not After": "2017-08-21 05:26:54",
            "Subject Hash": "97:95:14:4F",
            "Serial Number": "0D:FA",
            "SHA1 Finger Print": (
                "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D6"
            ),
            "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7",
            "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2",
            "Version": 1,
            "Key Size": 512,
            "Public Key": (
                "-----BEGIN PUBLIC"
                " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
                " PUBLIC KEY-----\n"
            ),
            "Issuer": {
                "C": "JP",
                "organizationName": "Frank4DD",
                "CN": "Frank4DD Web CA",
                "SP": "Tokyo",
                "L": "Chuo-ku",
                "emailAddress": "support@frank4dd.com",
                "OU": "WebCert Support",
            },
            "Issuer Hash": "92:DA:45:6B",
            "Not Before": "2012-08-22 05:26:54",
            "Subject": {
                "C": "JP",
                "SP": "Tokyo",
                "organizationName": "Frank4DD",
                "CN": "www.example.com",
            },
        }

        name = "keystore.jks"
        passphrase = "changeit"
        entries = [
            {
                "alias": "stringhost",
                "certificate": """-----BEGIN CERTIFICATE-----
                   MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
                   A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
                   MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
                   YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
                   ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
                   CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
                   ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
                   8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
                   AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
                   8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
                   2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
                   Hn+GmxZA\n-----END CERTIFICATE-----""",
            }
        ]

        test_return = {
            "name": name,
            "changes": {},
            "result": None,
            "comment": (
                "Alias stringhost would have been updated\nAlias oldhost would have"
                " been removed"
            ),
        }
        state_return = {
            "name": name,
            "changes": {"oldhost": "Removed", "stringhost": "Updated"},
            "result": True,
            "comment": "Alias stringhost updated.\nAlias oldhost removed.\n",
        }

        with patch.dict(keystore.__opts__, {"test": True}):
            with patch.dict(
                keystore.__salt__,
                {"keystore.list": MagicMock(return_value=cert_return)},
            ):
                with patch.dict(
                    keystore.__salt__,
                    {"x509.read_certificate": MagicMock(return_value=x509_return)},
                ):
                    self.assertDictEqual(
                        keystore.managed(name, passphrase, entries, force_remove=True),
                        test_return,
                    )

        with patch.dict(
            keystore.__salt__, {"keystore.list": MagicMock(return_value=cert_return)}
        ):
            with patch.dict(
                keystore.__salt__,
                {"x509.read_certificate": MagicMock(return_value=x509_return)},
            ):
                with patch.dict(
                    keystore.__salt__, {"keystore.remove": MagicMock(return_value=True)}
                ):
                    with patch.dict(
                        keystore.__salt__,
                        {"keystore.add": MagicMock(return_value=True)},
                    ):
                        self.assertDictEqual(
                            keystore.managed(
                                name, passphrase, entries, force_remove=True
                            ),
                            state_return,
                        )



import random

import pytest
import salt.modules.cmdmod
import salt.utils.platform
import salt.utils.win_lgpo_auditpol as win_lgpo_auditpol
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

settings = ["No Auditing", "Success", "Failure", "Success and Failure"]


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinLgpoAuditpolTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            win_lgpo_auditpol: {
                "__context__": {},
                "__salt__": {"cmd.run_all": salt.modules.cmdmod.run_all},
            }
        }

    def test_get_settings(self):
        names = win_lgpo_auditpol._get_valid_names()
        ret = win_lgpo_auditpol.get_settings(category="All")
        for name in names:
            self.assertIn(name, [k.lower() for k in ret])

    def test_get_settings_invalid_category(self):
        self.assertRaises(
            KeyError, win_lgpo_auditpol.get_settings, category="Fake Category"
        )

    @pytest.mark.slow_test
    def test_get_setting(self):
        names = win_lgpo_auditpol._get_valid_names()
        for name in names:
            ret = win_lgpo_auditpol.get_setting(name)
            self.assertIn(ret, settings)

    def test_get_setting_invalid_name(self):
        self.assertRaises(KeyError, win_lgpo_auditpol.get_setting, name="Fake Name")

    def test_set_setting(self):
        names = ["Credential Validation", "IPsec Driver", "File System", "SAM"]
        mock_set = MagicMock(return_value={"retcode": 0, "stdout": "Success"})
        with patch.object(salt.modules.cmdmod, "run_all", mock_set):
            with patch.object(
                win_lgpo_auditpol,
                "_get_valid_names",
                return_value=[k.lower() for k in names],
            ):
                for name in names:
                    value = random.choice(settings)
                    win_lgpo_auditpol.set_setting(name=name, value=value)
                    switches = win_lgpo_auditpol.settings[value]
                    cmd = 'auditpol /set /subcategory:"{}" {}'.format(name, switches)
                    mock_set.assert_called_once_with(cmd=cmd, python_shell=True)
                    mock_set.reset_mock()

    def test_set_setting_invalid_setting(self):
        names = ["Credential Validation", "IPsec Driver", "File System"]
        with patch.object(
            win_lgpo_auditpol,
            "_get_valid_names",
            return_value=[k.lower() for k in names],
        ):
            self.assertRaises(
                KeyError,
                win_lgpo_auditpol.set_setting,
                name="Fake Name",
                value="No Auditing",
            )

    def test_set_setting_invalid_value(self):
        names = ["Credential Validation", "IPsec Driver", "File System"]
        with patch.object(
            win_lgpo_auditpol,
            "_get_valid_names",
            return_value=[k.lower() for k in names],
        ):
            self.assertRaises(
                KeyError,
                win_lgpo_auditpol.set_setting,
                name="Credential Validation",
                value="Fake Value",
            )

    def test_get_auditpol_dump(self):
        names = win_lgpo_auditpol._get_valid_names()
        dump = win_lgpo_auditpol.get_auditpol_dump()
        for name in names:
            found = False
            for line in dump:
                if name.lower() in line.lower():
                    found = True
                    break
            self.assertTrue(found)



import pytest
from salt.cloud.clouds import ec2

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test__load_params_size():
    """
    Test that "Size" is added to params
    """
    kwargs = {"zone": "us-west-2", "size": 10}
    expected = {"Action": "CreateVolume", "AvailabilityZone": "us-west-2", "Size": 10}
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_snapshot():
    """
    Test that "SnapshotID" is added to params
    """
    kwargs = {"zone": "us-west-2", "snapshot": 1234}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
        "SnapshotId": 1234,
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_type():
    """
    Test that "VolumeType" is added to params
    """
    kwargs = {"zone": "us-west-2", "type": "gp2"}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
        "VolumeType": "gp2",
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_type_io1():
    """
    Test that "Iops" is added to params when type is io1
    """
    kwargs = {"zone": "us-west-2", "type": "io1", "iops": 1000}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
        "VolumeType": "io1",
        "Iops": 1000,
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_type_io2():
    """
    Test that "Iops" is added to params when type is io2
    """
    kwargs = {"zone": "us-west-2", "type": "io2", "iops": 1000}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
        "VolumeType": "io2",
        "Iops": 1000,
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_encrypted():
    """
    Test that "Encrypted" is added to params
    """
    kwargs = {"zone": "us-west-2", "encrypted": True}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
        "Encrypted": True,
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_encrypted_kmskeyid():
    """
    Test that "KmsKeyId" is added to params when passed with encrypted
    """
    kwargs = {"zone": "us-west-2", "encrypted": True, "kmskeyid": "keyid"}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
        "Encrypted": True,
        "KmsKeyId": "keyid",
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test__load_params_kmskeyid_without_encrypted():
    """
    Test that "KmsKeyId" is not added to params when encrypted is not passed
    """
    kwargs = {"zone": "us-west-2", "kmskeyid": "keyid"}
    expected = {
        "Action": "CreateVolume",
        "AvailabilityZone": "us-west-2",
    }
    result = ec2._load_params(kwargs=kwargs)
    assert result == expected


def test_create_volume_call_not_function():
    """
    Test check for function
    """
    ret = ec2.create_volume(kwargs={}, call="not a function")
    assert ret is False


def test_create_volume_no_zone():
    """
    Test that "zone" must be in kwargs
    """
    ret = ec2.create_volume(kwargs={}, call="function")
    assert ret is False


def test_create_volume_kmskeyid_without_encrypted():
    """
    Test that returns False if kmskeyid is passed without encrypted
    """
    kwargs = {"zone": "us-west-2", "kmskeyid": "keyid"}
    ret = ec2.create_volume(kwargs=kwargs, call="function")
    assert ret is False


def test_create_volume_missing_iops_io1():
    """
    Test that iops must be in kwargs if type is io1
    """
    kwargs = {"zone": "us-west-2", "type": "io1"}
    ret = ec2.create_volume(kwargs=kwargs, call="function")
    assert ret is False


def test_create_volume_missing_iops_io2():
    """
    Test that iops must be in kwargs if type is io2
    """
    kwargs = {"zone": "us-west-2", "type": "io1"}
    ret = ec2.create_volume(kwargs=kwargs, call="function")
    assert ret is False



"""
Idem Support
============

This util provides access to an idem-ready hub

.. versionadded:: 3002
"""
import logging
import sys

try:
    import pop.hub

    HAS_POP = True, None
except ImportError as e:
    HAS_POP = False, str(e)

log = logging.getLogger(__name__)

__virtualname__ = "idem"


def __virtual__():
    if sys.version_info < (3, 6):
        return False, "idem only works on python3.6 and later"
    if not HAS_POP[0]:
        return HAS_POP
    return __virtualname__


def hub():
    """
    Create a hub with idem ready to go and completely loaded
    """
    if "idem.hub" not in __context__:
        log.debug("Creating the POP hub")
        hub = pop.hub.Hub()

        log.debug("Initializing the loop")
        hub.pop.loop.create()

        log.debug("Loading subs onto hub")
        hub.pop.sub.add(dyne_name="config")
        # We aren't collecting grains at all but some exec modules depend on the sub being on the hub
        hub.pop.sub.add(dyne_name="grains")
        hub.pop.sub.add(dyne_name="idem")

        log.debug("Reading idem config options")
        hub.config.integrate.load(["acct", "idem"], "idem", parse_cli=False, logs=False)

        __context__["idem.hub"] = hub

    return __context__["idem.hub"]



import glob
import logging
import os
import re

import salt.utils.path
import salt.utils.yaml
from jinja2 import Environment, FileSystemLoader

log = logging.getLogger(__name__)


# Renders jinja from a template file
def render_jinja(_file, salt_data):
    j_env = Environment(loader=FileSystemLoader(os.path.dirname(_file)))
    j_env.globals.update(
        {
            "__opts__": salt_data["__opts__"],
            "__salt__": salt_data["__salt__"],
            "__grains__": salt_data["__grains__"],
            "__pillar__": salt_data["__pillar__"],
            "minion_id": salt_data["minion_id"],
        }
    )
    j_render = j_env.get_template(os.path.basename(_file)).render()
    return j_render


# Renders yaml from rendered jinja
def render_yaml(_file, salt_data):
    return salt.utils.yaml.safe_load(render_jinja(_file, salt_data))


# Returns a dict from a class yaml definition
def get_class(_class, salt_data):
    l_files = []
    saltclass_path = salt_data["path"]

    straight, sub_init, sub_straight = get_class_paths(_class, saltclass_path)

    for root, dirs, files in salt.utils.path.os_walk(
        os.path.join(saltclass_path, "classes"), followlinks=True
    ):
        for l_file in files:
            l_files.append(os.path.join(root, l_file))

    if straight in l_files:
        return render_yaml(straight, salt_data)

    if sub_straight in l_files:
        return render_yaml(sub_straight, salt_data)

    if sub_init in l_files:
        return render_yaml(sub_init, salt_data)

    log.warning("%s: Class definition not found", _class)
    return {}


def get_class_paths(_class, saltclass_path):
    """
    Converts the dotted notation of a saltclass class to its possible file counterparts.

    :param str _class: Dotted notation of the class
    :param str saltclass_path: Root to saltclass storage
    :return: 3-tuple of possible file counterparts
    :rtype: tuple(str)
    """
    straight = os.path.join(saltclass_path, "classes", "{}.yml".format(_class))
    sub_straight = os.path.join(
        saltclass_path, "classes", "{}.yml".format(_class.replace(".", os.sep))
    )
    sub_init = os.path.join(
        saltclass_path, "classes", _class.replace(".", os.sep), "init.yml"
    )
    return straight, sub_init, sub_straight


def get_class_from_file(_file, saltclass_path):
    """
    Converts the absolute path to a saltclass file back to the dotted notation.

    .. code-block:: python

       print(get_class_from_file('/srv/saltclass/classes/services/nginx/init.yml', '/srv/saltclass'))
       # services.nginx

    :param str _file: Absolute path to file
    :param str saltclass_path: Root to saltclass storage
    :return: class name in dotted notation
    :rtype: str
    """
    # remove classes path prefix
    _file = _file[len(os.path.join(saltclass_path, "classes")) + len(os.sep) :]
    # remove .yml extension
    _file = _file[:-4]
    # revert to dotted notation
    _file = _file.replace(os.sep, ".")
    # remove tailing init
    if _file.endswith(".init"):
        _file = _file[:-5]
    return _file


# Return environment
def get_env_from_dict(exp_dict_list):
    environment = ""
    for s_class in exp_dict_list:
        if "environment" in s_class:
            environment = s_class["environment"]
    return environment


# Merge dict b into a
def dict_merge(a, b, path=None):
    if path is None:
        path = []

    for key in b:
        if key in a:
            if isinstance(a[key], list) and isinstance(b[key], list):
                if b[key][0] == "^":
                    b[key].pop(0)
                    a[key] = b[key]
                else:
                    a[key].extend(b[key])
            elif isinstance(a[key], dict) and isinstance(b[key], dict):
                dict_merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


# Recursive search and replace in a dict
def dict_search_and_replace(d, old, new, expanded):
    for (k, v) in d.items():
        if isinstance(v, dict):
            dict_search_and_replace(d[k], old, new, expanded)

        if isinstance(v, list):
            x = 0
            for i in v:
                if isinstance(i, dict):
                    dict_search_and_replace(v[x], old, new, expanded)
                if isinstance(i, str):
                    if i == old:
                        v[x] = new
                x = x + 1

        if v == old:
            d[k] = new

    return d


# Retrieve original value from ${xx:yy:zz} to be expanded
def find_value_to_expand(x, v):
    a = x
    for i in v[2:-1].split(":"):
        if a is None:
            return v
        if i in a:
            a = a.get(i)
        else:
            return v
    return a


# Look for regexes and expand them
def find_and_process_re(_str, v, k, b, expanded):
    vre = re.finditer(r"(^|.)\$\{.*?\}", _str)
    if vre:
        for re_v in vre:
            re_str = str(re_v.group())
            if re_str.startswith("\\"):
                v_new = _str.replace(re_str, re_str.lstrip("\\"))
                b = dict_search_and_replace(b, _str, v_new, expanded)
                expanded.append(k)
            elif not re_str.startswith("$"):
                v_expanded = find_value_to_expand(b, re_str[1:])
                v_new = _str.replace(re_str[1:], v_expanded)
                b = dict_search_and_replace(b, _str, v_new, expanded)
                _str = v_new
                expanded.append(k)
            else:
                v_expanded = find_value_to_expand(b, re_str)
                if isinstance(v, str):
                    v_new = v.replace(re_str, v_expanded)
                else:
                    v_new = _str.replace(re_str, v_expanded)
                b = dict_search_and_replace(b, _str, v_new, expanded)
                _str = v_new
                v = v_new
                expanded.append(k)
    return b


# Return a dict that contains expanded variables if found
def expand_variables(a, b, expanded, path=None):
    if path is None:
        b = a.copy()
        path = []

    for (k, v) in a.items():
        if isinstance(v, dict):
            expand_variables(v, b, expanded, path + [str(k)])
        else:
            if isinstance(v, list):
                for i in v:
                    if isinstance(i, dict):
                        expand_variables(i, b, expanded, path + [str(k)])
                    if isinstance(i, str):
                        b = find_and_process_re(i, v, k, b, expanded)

            if isinstance(v, str):
                b = find_and_process_re(v, v, k, b, expanded)
    return b


def match_class_glob(_class, saltclass_path):
    """
    Takes a class name possibly including `*` or `?` wildcards (or any other wildcards supportet by `glob.glob`) and
    returns a list of expanded class names without wildcards.

    .. code-block:: python

       classes = match_class_glob('services.*', '/srv/saltclass')
       print(classes)
       # services.mariadb
       # services.nginx...


    :param str _class: dotted class name, globbing allowed.
    :param str saltclass_path: path to the saltclass root directory.

    :return: The list of expanded class matches.
    :rtype: list(str)
    """
    straight, sub_init, sub_straight = get_class_paths(_class, saltclass_path)
    classes = []
    matches = []
    matches.extend(glob.glob(straight))
    matches.extend(glob.glob(sub_straight))
    matches.extend(glob.glob(sub_init))
    if not matches:
        log.warning("%s: Class globbing did not yield any results", _class)
    for match in matches:
        classes.append(get_class_from_file(match, saltclass_path))
    return classes


def expand_classes_glob(classes, salt_data):
    """
    Expand the list of `classes` to no longer include any globbing.

    :param iterable(str) classes: Iterable of classes
    :param dict salt_data: configuration data
    :return: Expanded list of classes with resolved globbing
    :rtype: list(str)
    """
    all_classes = []
    expanded_classes = []
    saltclass_path = salt_data["path"]

    for _class in classes:
        all_classes.extend(match_class_glob(_class, saltclass_path))

    for _class in all_classes:
        if _class not in expanded_classes:
            expanded_classes.append(_class)

    return expanded_classes


def expand_classes_in_order(
    minion_dict, salt_data, seen_classes, expanded_classes, classes_to_expand
):
    # Get classes to expand from minion dictionary
    if not classes_to_expand and "classes" in minion_dict:
        classes_to_expand = minion_dict["classes"]

    classes_to_expand = expand_classes_glob(classes_to_expand, salt_data)

    # Now loop on list to recursively expand them
    for klass in classes_to_expand:
        if klass not in seen_classes:
            seen_classes.append(klass)
            expanded_classes[klass] = get_class(klass, salt_data)
            # Fix corner case where class is loaded but doesn't contain anything
            if expanded_classes[klass] is None:
                expanded_classes[klass] = {}

            # Merge newly found pillars into existing ones
            new_pillars = expanded_classes[klass].get("pillars", {})
            if new_pillars:
                dict_merge(salt_data["__pillar__"], new_pillars)

            # Now replace class element in classes_to_expand by expansion
            if expanded_classes[klass].get("classes"):
                l_id = classes_to_expand.index(klass)
                classes_to_expand[l_id:l_id] = expanded_classes[klass]["classes"]
                expand_classes_in_order(
                    minion_dict,
                    salt_data,
                    seen_classes,
                    expanded_classes,
                    classes_to_expand,
                )
            else:
                expand_classes_in_order(
                    minion_dict,
                    salt_data,
                    seen_classes,
                    expanded_classes,
                    classes_to_expand,
                )

    # We may have duplicates here and we want to remove them
    tmp = []
    for t_element in classes_to_expand:
        if t_element not in tmp:
            tmp.append(t_element)

    classes_to_expand = tmp

    # Now that we've retrieved every class in order,
    # let's return an ordered list of dicts
    ord_expanded_classes = []
    ord_expanded_states = []
    for ord_klass in classes_to_expand:
        ord_expanded_classes.append(expanded_classes[ord_klass])
        # And be smart and sort out states list
        # Address the corner case where states is empty in a class definition
        if (
            "states" in expanded_classes[ord_klass]
            and expanded_classes[ord_klass]["states"] is None
        ):
            expanded_classes[ord_klass]["states"] = {}

        if "states" in expanded_classes[ord_klass]:
            ord_expanded_states.extend(expanded_classes[ord_klass]["states"])

    # Add our minion dict as final element but check if we have states to process
    if "states" in minion_dict and minion_dict["states"] is None:
        minion_dict["states"] = []

    if "states" in minion_dict:
        ord_expanded_states.extend(minion_dict["states"])

    ord_expanded_classes.append(minion_dict)

    return ord_expanded_classes, classes_to_expand, ord_expanded_states


def expanded_dict_from_minion(minion_id, salt_data):
    _file = ""
    saltclass_path = salt_data["path"]
    # Start
    for root, dirs, files in salt.utils.path.os_walk(
        os.path.join(saltclass_path, "nodes"), followlinks=True
    ):
        for minion_file in files:
            if minion_file == "{}.yml".format(minion_id):
                _file = os.path.join(root, minion_file)

    # Load the minion_id definition if existing, else an empty dict
    node_dict = {}
    if _file:
        node_dict[minion_id] = render_yaml(_file, salt_data)
    else:
        log.warning("%s: Node definition not found", minion_id)
        node_dict[minion_id] = {}

    # Merge newly found pillars into existing ones
    dict_merge(salt_data["__pillar__"], node_dict[minion_id].get("pillars", {}))

    # Get 2 ordered lists:
    # expanded_classes: A list of all the dicts
    # classes_list: List of all the classes
    expanded_classes, classes_list, states_list = expand_classes_in_order(
        node_dict[minion_id], salt_data, [], {}, []
    )

    # Here merge the pillars together
    pillars_dict = {}
    for exp_dict in expanded_classes:
        if "pillars" in exp_dict:
            dict_merge(pillars_dict, exp_dict)

    return expanded_classes, pillars_dict, classes_list, states_list


def get_pillars(minion_id, salt_data):
    # Get 2 dicts and 2 lists
    # expanded_classes: Full list of expanded dicts
    # pillars_dict: dict containing merged pillars in order
    # classes_list: All classes processed in order
    # states_list: All states listed in order
    (
        expanded_classes,
        pillars_dict,
        classes_list,
        states_list,
    ) = expanded_dict_from_minion(minion_id, salt_data)

    # Retrieve environment
    environment = get_env_from_dict(expanded_classes)

    # Expand ${} variables in merged dict
    # pillars key shouldn't exist if we haven't found any minion_id ref
    if "pillars" in pillars_dict:
        pillars_dict_expanded = expand_variables(pillars_dict["pillars"], {}, [])
    else:
        pillars_dict_expanded = expand_variables({}, {}, [])

    # Build the final pillars dict
    pillars_dict = {}
    pillars_dict["__saltclass__"] = {}
    pillars_dict["__saltclass__"]["states"] = states_list
    pillars_dict["__saltclass__"]["classes"] = classes_list
    pillars_dict["__saltclass__"]["environment"] = environment
    pillars_dict["__saltclass__"]["nodename"] = minion_id
    pillars_dict.update(pillars_dict_expanded)

    return pillars_dict


def get_tops(minion_id, salt_data):
    # Get 2 dicts and 2 lists
    # expanded_classes: Full list of expanded dicts
    # pillars_dict: dict containing merged pillars in order
    # classes_list: All classes processed in order
    # states_list: All states listed in order
    (
        expanded_classes,
        pillars_dict,
        classes_list,
        states_list,
    ) = expanded_dict_from_minion(minion_id, salt_data)

    # Retrieve environment
    environment = get_env_from_dict(expanded_classes)

    # Build final top dict
    tops_dict = {}
    tops_dict[environment] = states_list

    return tops_dict



"""
unit tests for the Salt engines
"""

import logging

import pytest
import salt.config
import salt.engines as engines
import salt.utils.process
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class EngineTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.engine.sqs_events
    """

    def setup_loader_modules(self):
        return {engines: {}}

    @pytest.mark.slow_test
    def test_engine_module(self):
        """
        Test
        """
        mock_opts = salt.config.DEFAULT_MINION_OPTS.copy()
        mock_opts["__role"] = "minion"
        mock_opts["engines"] = [
            {"test_one": {"engine_module": "test"}},
            {"test_two": {"engine_module": "test"}},
        ]

        process_manager = salt.utils.process.ProcessManager()
        self.addCleanup(process_manager.stop_restarting)
        self.addCleanup(process_manager.kill_children)
        with patch.dict(engines.__opts__, mock_opts):
            salt.engines.start_engines(mock_opts, process_manager)
            process_map = process_manager._process_map
            count = 0
            for proc in process_map:
                count += 1
                fun = process_map[proc]["Process"].fun

                # Ensure function is start from the test engine
                self.assertEqual(fun, "test.start")

            # Ensure there were two engine started
            self.assertEqual(count, len(mock_opts["engines"]))



r"""
Salt Util for getting system information with the Performance Data Helper (pdh).
Counter information is gathered from current activity or log files.

Usage:

.. code-block:: python

    import salt.utils.win_pdh

    # Get a list of Counter objects
    salt.utils.win_pdh.list_objects()

    # Get a list of ``Processor`` instances
    salt.utils.win_pdh.list_instances('Processor')

    # Get a list of ``Processor`` counters
    salt.utils.win_pdh.list_counters('Processor')

    # Get the value of a single counter
    # \Processor(*)\% Processor Time
    salt.utils.win_pdh.get_counter('Processor', '*', '% Processor Time')

    # Get the values of multiple counters
    counter_list = [('Processor', '*', '% Processor Time'),
                    ('System', None, 'Context Switches/sec'),
                    ('Memory', None, 'Pages/sec'),
                    ('Server Work Queues', '*', 'Queue Length')]
    salt.utils.win_pdh.get_counters(counter_list)

    # Get all counters for the Processor object
    salt.utils.win_pdh.get_all_counters('Processor')
"""

# https://docs.microsoft.com/en-us/windows/desktop/perfctrs/using-the-pdh-functions-to-consume-counter-data

# https://www.cac.cornell.edu/wiki/index.php?title=Performance_Data_Helper_in_Python_with_win32pdh
import logging
import time

import salt.utils.platform
from salt.exceptions import CommandExecutionError

try:
    import pywintypes
    import win32pdh

    HAS_WINDOWS_MODULES = True
except ImportError:
    HAS_WINDOWS_MODULES = False


log = logging.getLogger(__file__)

# Define the virtual name
__virtualname__ = "pdh"


def __virtual__():
    """
    Only works on Windows systems with the PyWin32
    """
    if not salt.utils.platform.is_windows():
        return False, "salt.utils.win_pdh: Requires Windows"

    if not HAS_WINDOWS_MODULES:
        return False, "salt.utils.win_pdh: Missing required modules"

    return __virtualname__


class Counter:
    """
    Counter object
    Has enumerations and functions for working with counters
    """

    # The dwType field from GetCounterInfo returns the following, or'ed.
    # These come from WinPerf.h
    PERF_SIZE_DWORD = 0x00000000
    PERF_SIZE_LARGE = 0x00000100
    PERF_SIZE_ZERO = 0x00000200  # for Zero Length fields
    PERF_SIZE_VARIABLE_LEN = 0x00000300
    # length is in the CounterLength field of the Counter Definition structure

    # select one of the following values to indicate the counter field usage
    PERF_TYPE_NUMBER = 0x00000000  # a number (not a counter)
    PERF_TYPE_COUNTER = 0x00000400  # an increasing numeric value
    PERF_TYPE_TEXT = 0x00000800  # a text field
    PERF_TYPE_ZERO = 0x00000C00  # displays a zero

    # If the PERF_TYPE_NUMBER field was selected, then select one of the
    # following to describe the Number
    PERF_NUMBER_HEX = 0x00000000  # display as HEX value
    PERF_NUMBER_DECIMAL = 0x00010000  # display as a decimal integer
    PERF_NUMBER_DEC_1000 = 0x00020000  # display as a decimal/1000

    # If the PERF_TYPE_COUNTER value was selected then select one of the
    # following to indicate the type of counter
    PERF_COUNTER_VALUE = 0x00000000  # display counter value
    PERF_COUNTER_RATE = 0x00010000  # divide ctr / delta time
    PERF_COUNTER_FRACTION = 0x00020000  # divide ctr / base
    PERF_COUNTER_BASE = 0x00030000  # base value used in fractions
    PERF_COUNTER_ELAPSED = 0x00040000  # subtract counter from current time
    PERF_COUNTER_QUEUE_LEN = 0x00050000  # Use Queue len processing func.
    PERF_COUNTER_HISTOGRAM = 0x00060000  # Counter begins or ends a histogram

    # If the PERF_TYPE_TEXT value was selected, then select one of the
    # following to indicate the type of TEXT data.
    PERF_TEXT_UNICODE = 0x00000000  # type of text in text field
    PERF_TEXT_ASCII = 0x00010000  # ASCII using the CodePage field

    #  Timer SubTypes
    PERF_TIMER_TICK = 0x00000000  # use system perf. freq for base
    PERF_TIMER_100NS = 0x00100000  # use 100 NS timer time base units
    PERF_OBJECT_TIMER = 0x00200000  # use the object timer freq

    # Any types that have calculations performed can use one or more of the
    # following calculation modification flags listed here
    PERF_DELTA_COUNTER = 0x00400000  # compute difference first
    PERF_DELTA_BASE = 0x00800000  # compute base diff as well
    PERF_INVERSE_COUNTER = 0x01000000  # show as 1.00-value (assumes:
    PERF_MULTI_COUNTER = 0x02000000  # sum of multiple instances

    # Select one of the following values to indicate the display suffix (if any)
    PERF_DISPLAY_NO_SUFFIX = 0x00000000  # no suffix
    PERF_DISPLAY_PER_SEC = 0x10000000  # "/sec"
    PERF_DISPLAY_PERCENT = 0x20000000  # "%"
    PERF_DISPLAY_SECONDS = 0x30000000  # "secs"
    PERF_DISPLAY_NO_SHOW = 0x40000000  # value is not displayed

    def build_counter(obj, instance, instance_index, counter):
        r"""
        Makes a fully resolved counter path. Counter names are formatted like
        this:

        ``\Processor(*)\% Processor Time``

        The above breaks down like this:

            obj = 'Processor'
            instance = '*'
            counter = '% Processor Time'

        Args:

            obj (str):
                The top level object

            instance (str):
                The instance of the object

            instance_index (int):
                The index of the instance. Can usually be 0

            counter (str):
                The name of the counter

        Returns:
            Counter: A Counter object with the path if valid

        Raises:
            CommandExecutionError: If the path is invalid
        """
        path = win32pdh.MakeCounterPath(
            (None, obj, instance, None, instance_index, counter), 0
        )
        if win32pdh.ValidatePath(path) == 0:
            return Counter(path, obj, instance, instance_index, counter)
        raise CommandExecutionError("Invalid counter specified: {}".format(path))

    build_counter = staticmethod(build_counter)

    def __init__(self, path, obj, instance, index, counter):
        self.path = path
        self.obj = obj
        self.instance = instance
        self.index = index
        self.counter = counter
        self.handle = None
        self.info = None
        self.type = None

    def add_to_query(self, query):
        """
        Add the current path to the query

        Args:
            query (obj):
                The handle to the query to add the counter
        """
        self.handle = win32pdh.AddCounter(query, self.path)

    def get_info(self):
        """
        Get information about the counter

        .. note::
            GetCounterInfo sometimes crashes in the wrapper code. Fewer crashes
            if this is called after sampling data.
        """
        if not self.info:
            ci = win32pdh.GetCounterInfo(self.handle, 0)
            self.info = {
                "type": ci[0],
                "version": ci[1],
                "scale": ci[2],
                "default_scale": ci[3],
                "user_data": ci[4],
                "query_user_data": ci[5],
                "full_path": ci[6],
                "machine_name": ci[7][0],
                "object_name": ci[7][1],
                "instance_name": ci[7][2],
                "parent_instance": ci[7][3],
                "instance_index": ci[7][4],
                "counter_name": ci[7][5],
                "explain_text": ci[8],
            }
        return self.info

    def value(self):
        """
        Return the counter value

        Returns:
            long: The counter value
        """
        (counter_type, value) = win32pdh.GetFormattedCounterValue(
            self.handle, win32pdh.PDH_FMT_DOUBLE
        )
        self.type = counter_type
        return value

    def type_string(self):
        """
        Returns the names of the flags that are set in the Type field

        It can be used to format the counter.
        """
        type = self.get_info()["type"]
        type_list = []
        for member in dir(self):
            if member.startswith("PERF_"):
                bit = getattr(self, member)
                if bit and bit & type:
                    type_list.append(member[5:])
        return type_list

    def __str__(self):
        return self.path


def list_objects():
    """
    Get a list of available counter objects on the system

    Returns:
        list: A list of counter objects
    """
    return sorted(win32pdh.EnumObjects(None, None, -1, 0))


def list_counters(obj):
    """
    Get a list of counters available for the object

    Args:
        obj (str):
            The name of the counter object. You can get a list of valid names
            using the ``list_objects`` function

    Returns:
        list: A list of counters available to the passed object
    """
    return win32pdh.EnumObjectItems(None, None, obj, -1, 0)[0]


def list_instances(obj):
    """
    Get a list of instances available for the object

    Args:
        obj (str):
            The name of the counter object. You can get a list of valid names
            using the ``list_objects`` function

    Returns:
        list: A list of instances available to the passed object
    """
    return win32pdh.EnumObjectItems(None, None, obj, -1, 0)[1]


def build_counter_list(counter_list):
    r"""
    Create a list of Counter objects to be used in the pdh query

    Args:
        counter_list (list):
            A list of tuples containing counter information. Each tuple should
            contain the object, instance, and counter name. For example, to
            get the ``% Processor Time`` counter for all Processors on the
            system (``\Processor(*)\% Processor Time``) you would pass a tuple
            like this:

            ```
            counter_list = [('Processor', '*', '% Processor Time')]
            ```

            If there is no ``instance`` for the counter, pass ``None``

            Multiple counters can be passed like so:

            ```
            counter_list = [('Processor', '*', '% Processor Time'),
                            ('System', None, 'Context Switches/sec')]
            ```

            .. note::
                Invalid counters are ignored

    Returns:
        list: A list of Counter objects
    """
    counters = []
    index = 0
    for obj, instance, counter_name in counter_list:
        try:
            counter = Counter.build_counter(obj, instance, index, counter_name)
            index += 1
            counters.append(counter)
        except CommandExecutionError as exc:
            # Not a valid counter
            log.debug(exc.strerror)
            continue
    return counters


def get_all_counters(obj, instance_list=None):
    """
    Get the values for all counters available to a Counter object

    Args:

        obj (str):
            The name of the counter object. You can get a list of valid names
            using the ``list_objects`` function

        instance_list (list):
            A list of instances to return. Use this to narrow down the counters
            that are returned.

            .. note::
                ``_Total`` is returned as ``*``
    """
    counters, instances_avail = win32pdh.EnumObjectItems(None, None, obj, -1, 0)

    if instance_list is None:
        instance_list = instances_avail

    if not isinstance(instance_list, list):
        instance_list = [instance_list]

    counter_list = []
    for counter in counters:
        for instance in instance_list:
            instance = "*" if instance.lower() == "_total" else instance
            counter_list.append((obj, instance, counter))
        else:  # pylint: disable=useless-else-on-loop
            counter_list.append((obj, None, counter))

    return get_counters(counter_list) if counter_list else {}


def get_counters(counter_list):
    """
    Get the values for the passes list of counters

    Args:
        counter_list (list):
            A list of counters to lookup

    Returns:
        dict: A dictionary of counters and their values
    """
    if not isinstance(counter_list, list):
        raise CommandExecutionError("counter_list must be a list of tuples")

    try:
        # Start a Query instances
        query = win32pdh.OpenQuery()

        # Build the counters
        counters = build_counter_list(counter_list)

        # Add counters to the Query
        for counter in counters:
            counter.add_to_query(query)

        # https://docs.microsoft.com/en-us/windows/desktop/perfctrs/collecting-performance-data
        win32pdh.CollectQueryData(query)
        # The sleep here is required for counters that require more than 1
        # reading
        time.sleep(1)
        win32pdh.CollectQueryData(query)
        ret = {}

        for counter in counters:
            try:
                ret.update({counter.path: counter.value()})
            except pywintypes.error as exc:
                if exc.strerror == "No data to return.":
                    # Some counters are not active and will throw an error if
                    # there is no data to return
                    continue
                else:
                    raise

    except pywintypes.error as exc:
        if exc.strerror == "No data to return.":
            # Sometimess, win32pdh.CollectQueryData can err
            # so just ignore it
            return {}
        else:
            raise

    finally:
        win32pdh.CloseQuery(query)

    return ret


def get_counter(obj, instance, counter):
    """
    Get the value of a single counter

    Args:

        obj (str):
            The name of the counter object. You can get a list of valid names
            using the ``list_objects`` function

        instance (str):
            The counter instance you wish to return. Get a list of instances
            using the ``list_instances`` function

            .. note::
                ``_Total`` is returned as ``*``

        counter (str):
            The name of the counter. Get a list of counters using the
            ``list_counters`` function
    """
    return get_counters([(obj, instance, counter)])



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import pytest
import salt.states.win_system as win_system
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_system: {}}


def test_computer_desc():
    """
    Test to manage the computer's description field
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(side_effect=["salt", "stack", "stack"])
    with patch.dict(win_system.__salt__, {"system.get_computer_desc": mock}):
        ret.update({"comment": "Computer description already set to 'salt'"})
        assert win_system.computer_desc("salt") == ret

        with patch.dict(win_system.__opts__, {"test": True}):
            ret.update(
                {
                    "result": None,
                    "comment": "Computer description will be changed to 'salt'",
                }
            )
            assert win_system.computer_desc("salt") == ret

        with patch.dict(win_system.__opts__, {"test": False}):
            mock = MagicMock(return_value={"Computer Description": "nfs"})
            with patch.dict(win_system.__salt__, {"system.set_computer_desc": mock}):
                ret.update(
                    {
                        "result": False,
                        "comment": "Unable to set computer description to 'salt'",
                    }
                )
                assert win_system.computer_desc("salt") == ret


def test_computer_name():
    """
    Test to manage the computer's name
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}
    mock = MagicMock(return_value="salt")
    with patch.dict(win_system.__salt__, {"system.get_computer_name": mock}):
        mock = MagicMock(side_effect=[None, "SALT", "Stack", "stack"])
        with patch.dict(
            win_system.__salt__, {"system.get_pending_computer_name": mock}
        ):
            ret.update({"comment": "Computer name already set to 'salt'"})
            assert win_system.computer_name("salt") == ret

            ret.update(
                {
                    "comment": (
                        "The current computer name"
                        " is 'salt', but will be changed to 'salt' on"
                        " the next reboot"
                    )
                }
            )
            assert win_system.computer_name("salt") == ret

            with patch.dict(win_system.__opts__, {"test": True}):
                ret.update(
                    {
                        "result": None,
                        "comment": "Computer name will be changed to 'salt'",
                    }
                )
                assert win_system.computer_name("salt") == ret

            with patch.dict(win_system.__opts__, {"test": False}):
                mock = MagicMock(return_value=False)
                with patch.dict(
                    win_system.__salt__, {"system.set_computer_name": mock}
                ):
                    ret.update(
                        {
                            "comment": "Unable to set computer name to 'salt'",
                            "result": False,
                        }
                    )
                    assert win_system.computer_name("salt") == ret


def test_hostname():
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(return_value="minion")
    with patch.dict(win_system.__salt__, {"system.get_hostname": mock}):
        mock = MagicMock(return_value=True)
        with patch.dict(win_system.__salt__, {"system.set_hostname": mock}):
            ret.update(
                {
                    "comment": (
                        "The current hostname is 'minion', "
                        "but will be changed to 'salt' on the next reboot"
                    ),
                    "changes": {"hostname": "salt"},
                }
            )
            assert win_system.hostname("salt") == ret

        mock = MagicMock(return_value=False)
        with patch.dict(win_system.__salt__, {"system.set_hostname": mock}):
            ret.update(
                {"comment": "Unable to set hostname", "changes": {}, "result": False}
            )
            assert win_system.hostname("salt") == ret

    mock = MagicMock(return_value="salt")
    with patch.dict(win_system.__salt__, {"system.get_hostname": mock}):
        ret.update(
            {
                "comment": "Hostname is already set to 'salt'",
                "changes": {},
                "result": True,
            }
        )

        assert win_system.hostname("salt") == ret

    mock = MagicMock(return_value="salt")
    with patch.dict(win_system.__salt__, {"system.get_hostname": mock}):
        ret.update(
            {
                "name": "SALT",
                "comment": "Hostname is already set to 'SALT'",
                "changes": {},
                "result": True,
            }
        )

        assert win_system.hostname("SALT") == ret


def test_workgroup():
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(return_value={"Workgroup": "WORKGROUP"})
    with patch.dict(win_system.__salt__, {"system.get_domain_workgroup": mock}):
        mock = MagicMock(return_value=True)
        with patch.dict(win_system.__salt__, {"system.set_domain_workgroup": mock}):
            with patch.dict(win_system.__opts__, {"test": True}):
                ret.update(
                    {
                        "comment": "Computer will be joined to workgroup 'WERKGROUP'",
                        "name": "WERKGROUP",
                        "result": None,
                    }
                )
                assert win_system.workgroup("WERKGROUP") == ret

    mock = MagicMock(return_value={"Workgroup": "WORKGROUP"})
    with patch.dict(win_system.__salt__, {"system.get_domain_workgroup": mock}):
        mock = MagicMock(return_value=True)
        with patch.dict(win_system.__salt__, {"system.set_domain_workgroup": mock}):
            with patch.dict(win_system.__opts__, {"test": False}):
                ret.update(
                    {
                        "comment": "Workgroup is already set to 'WORKGROUP'",
                        "name": "WORKGROUP",
                        "result": True,
                    }
                )
                assert win_system.workgroup("WORKGROUP") == ret

    mock = MagicMock(
        side_effect=[{"Workgroup": "WORKGROUP"}, {"Workgroup": "WERKGROUP"}]
    )
    with patch.dict(win_system.__salt__, {"system.get_domain_workgroup": mock}):
        mock = MagicMock(return_value=True)
        with patch.dict(win_system.__salt__, {"system.set_domain_workgroup": mock}):
            with patch.dict(win_system.__opts__, {"test": False}):
                ret.update(
                    {
                        "comment": (
                            "The workgroup has been changed from 'WORKGROUP' to"
                            " 'WERKGROUP'"
                        ),
                        "name": "WERKGROUP",
                        "changes": {"new": "WERKGROUP", "old": "WORKGROUP"},
                        "result": True,
                    }
                )
                assert win_system.workgroup("WERKGROUP") == ret



"""
Botocore waiters for elasticsearch that are not present in boto3+botocore (yet).

:codeauthor: Herbert Buurman <herbert.buurman@ogd.nl>
:depends: boto3
"""

import salt.utils.versions
from salt.exceptions import SaltInvocationError

try:
    import botocore.waiter
except ImportError:
    pass


WAITER_CONFIGS = {
    "ESDomainAvailable": {
        "delay": 60,
        "operation": "DescribeElasticsearchDomainConfig",
        "maxAttempts": 60,
        "acceptors": [
            {
                "expected": "Active",
                "matcher": "path",
                "state": "success",
                "argument": "DomainConfig.ElasticsearchClusterConfig.Status.State",
            },
            {
                "expected": True,
                "matcher": "pathAny",
                "state": "failure",
                "argument": "DomainConfig.*.Status.PendingDeletion",
            },
        ],
    },
    "ESUpgradeFinished": {
        "delay": 60,
        "operation": "DescribeElasticsearchDomain",
        "maxAttempts": 60,
        "acceptors": [
            {
                "expected": False,
                "matcher": "path",
                "state": "success",
                "argument": "DomainStatus.UpgradeProcessing",
            }
        ],
    },
    "ESDomainDeleted": {
        "delay": 30,
        "operation": "DescribeElasticsearchDomain",
        "maxAttempts": 60,
        "acceptors": [
            {
                "expected": True,
                "matcher": "path",
                "state": "retry",
                "argument": "DomainStatus.Deleted",
            },
            {
                "expected": False,
                "matcher": "path",
                "state": "failure",
                "argument": "DomainStatus.Processing",
            },
            {
                "expected": "ResourceNotFoundException",
                "matcher": "error",
                "state": "success",
            },
        ],
    },
    "ESDomainCreated": {
        "delay": 30,
        "operation": "DescribeElasticsearchDomain",
        "maxAttempts": 60,
        "acceptors": [
            {
                "expected": True,
                "matcher": "path",
                "state": "success",
                "argument": "DomainStatus.Created",
            }
        ],
    },
}


def __virtual__():
    """
    Only load if botocore libraries exist.
    """
    return salt.utils.versions.check_boto_reqs(check_boto=False)


@salt.utils.decorators.is_deprecated(globals(), "Sulfur")
def get_waiter(client, waiter=None, waiter_config=None):
    """
    Gets a botocore waiter using either one of the preconfigured models by name
    ``waiter``, or with a manually supplied ``waiter_config``.

    :param botoclient client: The botocore client to use.
    :param str waiter: The name of the waiter config to use.
        Either ``waiter`` or ``waiter_config`` must be supplied.
        If both ``waiter`` and ``waiter_config`` are supplied, ``waiter`` takes
        presedence, unless no configuration for ``waiter`` exists.
    :param dict waiter_config: The manual waiter config to use.
        Either waiter or waiter_config must be supplied.

    :returns botocore.waiter
    """
    if not any((waiter, waiter_config)):
        raise SaltInvocationError(
            "At least one of waiter or waiter_config must be specified."
        )
    waiter_model = botocore.waiter.WaiterModel(
        {"version": 2, "waiters": {waiter: WAITER_CONFIGS.get(waiter, waiter_config)}}
    )
    return botocore.waiter.create_waiter_with_client(waiter, waiter_model, client)


@salt.utils.decorators.is_deprecated(globals(), "Sulfur")
def list_waiters():
    """
    Lists the builtin waiter configuration names.

    :returns list
    """
    return WAITER_CONFIGS.keys()



import os
import time

import pytest
import salt.utils.files
import salt.utils.platform
from salt.beacons import watchdog

pytestmark = [
    pytest.mark.skipif(
        watchdog.HAS_WATCHDOG is False, reason="watchdog is not available"
    ),
    pytest.mark.skipif(
        salt.utils.platform.is_darwin(),
        reason="Tests were being skipped pre macos under nox. Keep it like that for now.",
    ),
]


def check_events(config):
    total_delay = 1
    delay_per_loop = 20e-3

    for _ in range(int(total_delay / delay_per_loop)):
        events = watchdog.beacon(config)

        if events:
            return events

        time.sleep(delay_per_loop)

    return []


def create(path, content=None):
    with salt.utils.files.fopen(path, "w") as f:
        if content:
            f.write(content)
        os.fsync(f)


@pytest.fixture
def configure_loader_modules():
    return {watchdog: {}}


def assertValid(config):
    ret = watchdog.validate(config)
    assert ret == (True, "Valid beacon configuration")


def test_empty_config():
    config = [{}]
    ret = watchdog.beacon(config)
    assert ret == []


@pytest.mark.skipif(
    salt.utils.platform.is_freebsd(),
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_file_create(tmp_path):
    path = str(tmp_path / "tmpfile")

    config = [{"directories": {str(tmp_path): {"mask": ["create"]}}}]
    assertValid(config)
    assert watchdog.beacon(config) == []

    create(path)

    ret = check_events(config)
    assert len(ret) == 1
    assert ret[0]["path"] == path
    assert ret[0]["change"] == "created"


def test_file_modified(tmp_path):
    path = str(tmp_path / "tmpfile")
    # Create triggers a modify event along with the create event in Py3
    # So, let's do this before configuring the beacon
    create(path)

    config = [{"directories": {str(tmp_path): {"mask": ["modify"]}}}]
    assertValid(config)
    assert watchdog.beacon(config) == []

    create(path, "some content")

    ret = check_events(config)

    modified = False
    for event in ret:
        # "modified" requires special handling
        # A modification sometimes triggers 2 modified events depending on
        # the OS and the python version
        # When the "modified" event triggers on modify, it will have the
        # path to the temp file (path), other modified events will contain
        # the path minus "tmpfile" and will not match. That's how we'll
        # distinguish the two
        if event["change"] == "modified":
            if event["path"] == path:
                modified = True

    # Check results of the for loop to validate modified
    assert modified


def test_file_deleted(tmp_path):
    path = str(tmp_path / "tmpfile")
    create(path)

    config = [{"directories": {str(tmp_path): {"mask": ["delete"]}}}]
    assertValid(config)
    assert watchdog.beacon(config) == []

    os.remove(path)

    ret = check_events(config)
    assert len(ret) == 1
    assert ret[0]["path"] == path
    assert ret[0]["change"] == "deleted"


@pytest.mark.skipif(
    salt.utils.platform.is_freebsd(),
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_file_moved(tmp_path):
    path = str(tmp_path / "tmpfile")
    create(path)

    config = [{"directories": {str(tmp_path): {"mask": ["move"]}}}]
    assertValid(config)
    assert watchdog.beacon(config) == []

    os.rename(path, path + "_moved")

    ret = check_events(config)
    assert len(ret) == 1
    assert ret[0]["path"] == path
    assert ret[0]["change"] == "moved"


@pytest.mark.skipif(
    salt.utils.platform.is_freebsd(),
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
def test_file_create_in_directory(tmp_path):
    config = [{"directories": {str(tmp_path): {"mask": ["create"]}}}]
    assertValid(config)
    assert watchdog.beacon(config) == []

    path = str(tmp_path / "tmpfile")
    create(path)

    ret = check_events(config)
    assert len(ret) == 1
    assert ret[0]["path"] == path
    assert ret[0]["change"] == "created"


@pytest.mark.skipif(
    salt.utils.platform.is_freebsd(),
    reason="Skip on FreeBSD - does not yet have full inotify/watchdog support",
)
@pytest.mark.slow_test
def test_trigger_all_possible_events(tmp_path):
    path = str(tmp_path / "tmpfile")
    moved = path + "_moved"

    config = [{"directories": {str(tmp_path): {}}}]
    assertValid(config)
    assert watchdog.beacon(config) == []

    # create
    create(path)
    # modify
    create(path, "modified content")
    # move
    os.rename(path, moved)
    # delete
    os.remove(moved)

    # Give the events time to load into the queue
    time.sleep(1)

    ret = check_events(config)

    events = {"created": "", "deleted": "", "moved": ""}
    modified = False
    for event in ret:
        if event["change"] == "created":
            assert event["path"] == path
            events.pop("created", "")
        if event["change"] == "moved":
            assert event["path"] == path
            events.pop("moved", "")
        if event["change"] == "deleted":
            assert event["path"] == moved
            events.pop("deleted", "")
        # "modified" requires special handling
        # All events [created, moved, deleted] also trigger a "modified"
        # event on Linux
        # Only the "created" event triggers a modified event on Py3 Windows
        # When the "modified" event triggers on modify, it will have the
        # path to the temp file (path), other modified events will contain
        # the path minus "tmpfile" and will not match. That's how we'll
        # distinguish the two
        if event["change"] == "modified":
            if event["path"] == path:
                modified = True

    # Check results of the for loop to validate modified
    assert modified

    # Make sure all events were checked
    assert events == {}
